from __future__ import absolute_import
from __future__ import unicode_literals
from flask import redirect, url_for, render_template, make_response, request,\
        json, session, current_app
import flask_babel
from flask_login import login_required, current_user
from babel import get_locale_identifier, negotiate_locale, parse_locale
import six
from .. import db, babel, sentry
from ..models import Request, ActionType
from ..auth import PermissionType
from ..auth.models import Permission, Division
from ..util import jsonify, varies, locale


if six.PY3:
    unicode = str


@login_required
def index():
    """The index page for EVE-SRP."""
    return redirect(url_for('requests.personal_requests'))


@varies('Accept', 'X-Requested-With')
def error_page(error):
    """View function for displaying error pages."""
    # Try to get some meaningful bits of information about the error
    # HTTPExceptions (raised by abort()) have code and description attributes.
    # Try to prefer those rich values over more generic information.
    code = error.code if hasattr(error, 'code') else 500
    if hasattr(error, 'description'):
        description = error.description
    else:
        description = str(error)
    name = error.name if hasattr(error, 'name') else u'Application Error'
    # Give the error information in a machine readable format for APIs
    if request.is_json or request.is_xhr:
        response_content = jsonify(description=description, code=code)
    elif request.is_xml:
        pass
    else:
        response_content = render_template('error.html', code=code,
                description=description, name=name, title=code)
    if code == 500 and current_app.config.get('SENTRY_RELEASE') is not None:
        sentry.captureException()
    # Give a default response code for generic exceptions
    return make_response(response_content, code)


def request_count(permission, statuses=None):
    """Function intended for counting the number of requests for Jinja
    templates.
    """
    if statuses is None:
        if permission == PermissionType.review:
            statuses = (ActionType.evaluating,)
        elif permission == PermissionType.pay:
            statuses = (ActionType.approved,)
        elif permission == PermissionType.submit:
            statuses = (ActionType.incomplete,)
        else:
            return 0
    elif statuses in ActionType.statuses:
        statuses = (statuses,)
    permissions = current_user.permissions.\
            filter(Permission.permission==permission).\
            subquery()
    divisions = db.session.query(Division.id).\
            join(permissions).\
            subquery()
    requests = db.session.query(db.func.count(db.distinct(Request.id))).\
            join(divisions).\
            filter(Request.status.in_(statuses))
    if permission == PermissionType.submit:
        requests = requests.filter(Request.submitter==current_user)
    return requests.one()[0]


def update_navbar(response):
    if request.endpoint == 'static':
        return response
    if 'application/json' not in response.mimetype:
        return response
    response_json = json.loads(response.get_data())
    counts = {
        'pending': 0,
        'payouts': 0,
        'personal': 0
    }
    response_json['nav_counts'] = counts
    # Unauthenticated users get nothing
    if not current_user.is_authenticated:
        response.set_data(json.dumps(response_json))
        return response
    counts['pending'] = request_count(PermissionType.review)
    counts['payouts'] = request_count(PermissionType.pay)
    counts['personal'] = request_count(PermissionType.submit)
    response.set_data(json.dumps(response_json))
    return response


def detect_language():
    if 'lang' in request.args:
        requested_locale = request.args['lang']
        locales = [unicode(l) for l in babel.list_translations()]
        locale = negotiate_locale([requested_locale,], locales)
        session['locale'] = locale
        flask_babel.refresh()


def locale_selector():
    requested_locale = session.get('locale')
    supported_locales = [unicode(l) for l in locale.enabled_locales()]
    if requested_locale is not None and \
            requested_locale not in supported_locales:
        requested_locale = None
        del session['locale']
    return requested_locale



"""Add a Full Text Search index for MySQL

Revision ID: 19506187e7aa
Revises: 108d7a5881c7
Create Date: 2014-07-23 11:32:21.748067

"""

# revision identifiers, used by Alembic.
revision = '19506187e7aa'
down_revision = '108d7a5881c7'

from alembic import op
import sqlalchemy as sa


def upgrade():
    bind = op.get_bind()
    if bind.dialect.name == 'mysql':
        op.execute('CREATE FULLTEXT INDEX ix_request_details_fulltext ON '
                   'request (details);')


def downgrade():
    bind = op.get_bind()
    if bind.dialect.name == 'mysql':
        op.drop_index('ix_request_details_fulltext', 'details')



from __future__ import absolute_import
from flask import url_for, render_template, redirect, abort, flash, request,\
        Blueprint, current_app
from flask_babel import gettext, lazy_gettext
from flask_login import login_required, fresh_login_required, current_user
from flask_wtf import Form
import six
from six.moves import map
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from wtforms.fields import StringField, SubmitField, HiddenField, SelectField,\
        Label
from wtforms.validators import InputRequired, AnyOf, NumberRange

from ..models import db
from ..auth import PermissionType
from ..auth.models import Division, Permission, Entity
from ..util import jsonify, varies


blueprint = Blueprint('divisions', __name__)


@blueprint.route('/')
@fresh_login_required
def permissions():
    """Show a page listing all divisions.
    """
    if current_user.admin:
        return render_template('divisions.html',
                divisions=Division.query.all())
    if current_user.has_permission(PermissionType.admin):
        admin_permissions = current_user.permissions.filter_by(
                permission=PermissionType.admin).values(Permission.division_id)
        admin_permissions = map(lambda x: x[0], admin_permissions)
        divisions = db.session.query(Division).\
                filter(Division.id.in_(admin_permissions))
        return render_template('divisions.html', divisions=divisions)
    return render_template('permissions.html')
    return abort(403)


class AddDivisionForm(Form):
    # TRANS: On a form for creating a new division, this is a label for the
    # TRANS: name of the division.
    name = StringField(lazy_gettext(u'Division Name'),
            validators=[InputRequired()])

    # TRANS: On a form for creating a new division, this is a button for
    # TRANS: creating a new division (by submitting the form).
    submit = SubmitField(lazy_gettext(u'Create Division'))


@blueprint.route('/add/', methods=['GET', 'POST'])
@fresh_login_required
def add_division():
    """Present a form for adding a division and also process that form.

    Only accesible to adminstrators.
    """
    if not current_user.admin:
        return abort(403)
    form = AddDivisionForm()
    if form.validate_on_submit():
        division = Division(form.name.data)
        db.session.add(division)
        db.session.commit()
        return redirect(url_for('.get_division_details',
            division_id=division.id))
    return render_template('form.html', form=form,
            # TRANS: The title for a page for creating new divisions.
            title=gettext(u'Create Division'))


class ChangeEntity(Form):
    form_id = HiddenField(default='entity')
    id_ = HiddenField()
    name = StringField()
    permission = HiddenField(validators=[AnyOf(list(PermissionType.values()))])
    action = HiddenField(validators=[AnyOf(('add', 'delete'))])


#: List of tuples enumerating attributes that can be transformed/linked.
#: Mainly used as the choices argument to :py:class:`~.SelectField`
transformer_choices = [
    ('', u''),
    # TRANS: Label for fields showing the name of a pilot.
    ('pilot', lazy_gettext(u'Pilot')),
    # TRANS: Label for the corporation a pilot is in.
    ('corporation', lazy_gettext(u'Corporation')),
    # TRANS: Label for the alliance a pilot is in.
    ('alliance', lazy_gettext(u'Alliance')),
    # TRANS: Label for the solar system a loss occured in.
    ('system', lazy_gettext(u'Solar System')),
    # TRANS: Label for the constellation a loss occured in.
    ('constellation', lazy_gettext(u'Constellation')),
    # TRANS: Label for the region a loss occured in.
    ('region', lazy_gettext(u'Region')),
    # TRANS: Label for the type of ship that was lost.
    ('ship_type', lazy_gettext(u'Ship')),
    # TRANS: Label for the status a request is in (ex: Unevaluated, Approved)
    ('status', lazy_gettext(u'Request Status')),
]


class ChangeTransformer(Form):
    form_id = HiddenField(default='transformer')

    # TRANS: The a label for a selection field for selecting which attribute
    # TRANS: to transform. See the translation for 'Attribute Transformer'.
    attribute = SelectField(lazy_gettext(u'Attribute'),
            choices=transformer_choices)

    # TRANS: The label for a selection field for selecting the transformer for
    # TRANS: an attribute. See the translation for 'Attribute Transformer'.
    transformer = SelectField(lazy_gettext(u'Transformer'), choices=[])


def transformer_choices(attr):
    """Conveniece function for generating a list of transformer option tuples.

    :param attr str: the name of the attribute to make a list for.
    :return: A list of tuples suitable for the choices argument of\
        :py:class:`StringField`
    :rtype: list
    """
    default_transformers = [
        ('none', 'None'),
    ]
    choices = default_transformers
    if attr in current_app.url_transformers:
        for transformer in current_app.url_transformers[attr]:
            choices.append((transformer, transformer))
    return choices


@blueprint.route('/<int:division_id>/', methods=['GET'])
@fresh_login_required
@varies('Accept', 'X-Requested-With')
def get_division_details(division_id=None, division=None):
    """Generate a page showing the details of a division.

    Shows which groups and individuals have been granted permissions to each
    division.

    Only accesible to administrators.

    :param int division_id: The ID number of the division
    """
    if division is None:
        division = Division.query.get_or_404(division_id)
    if not current_user.admin and not \
            current_user.has_permission(PermissionType.admin, division):
        abort(403)
    if request.is_json or request.is_xhr:
        return jsonify(division._json(True))
    return render_template(
            'division_detail.html',
            division=division,
            entity_form=ChangeEntity(formdata=None),
            transformer_form=ChangeTransformer(formdata=None),
    )


def _modify_division_entity(division):
    """Handle POST requests for adding/removing entities form a Division."""
    form = ChangeEntity()
    if form.validate():
        entity = None
        if form.id_.data != '':
            current_app.logger.debug("Looking up entity by ID: {}".format(
                form.id_.data))
            entity = Entity.query.get(form.id_.data)
            if entity is None:
                # TRANS: This is an error message when there's a problem 
                # TRANS: granting a permission to a user or group
                # TRANS: (collectively called 'entities'). The '#' is not
                # TRANS: special, but the '%s(in_num)d' will be replaced with
                # TRANS: the ID number that was attempted to be added.
                flash(gettext(u"No entity with ID #%(id_num)d.",
                        id_num=form.id_.data),
                    category=u'error')
        else:
            current_app.logger.debug(u"Looking up entity by name: '{}'"\
                    .format(form.name.data))
            try:
                entity = Entity.query.filter_by(
                        name=form.name.data).one()
            except NoResultFound:
                # TRANS: Error message when a user or group with a given name
                # TRANS: cannot be found.
                flash(gettext(u"No entities with the name '%(name)s' found.",
                        name=form.name.data),
                    category=u'error')
            except MultipleResultsFound:
                # TRANS: Error message when multiple users and/or groups are
                # TRANS: found with a given name.
                flash(gettext(
                        u"Multiple entities with the name '%(name)s' found.",
                        name=form.name.data),
                    category=u'error')
            else:
                current_app.logger.debug("entity lookup success")
        if entity is None:
            return get_division_details(division=division), 404, None
        # The entity has been found, create the query for the requested
        # Permission.
        permission_type = PermissionType.from_string(
                form.permission.data)
        permission_query = Permission.query.filter_by(
                division=division,
                entity=entity,
                permission=permission_type)
        # The response for both add and delete actions depends on whether the
        # Permission is found, so look it up first.
        try:
            permission = permission_query.one()
        except NoResultFound:
            if form.action.data == 'add':
                db.session.add(
                    Permission(division, permission_type, entity))
                # TRANS: Message show when granting a permission to a user or
                # TRANS: group.
                flash(gettext(u"%(name)s is now a %(role)s.",
                        name=entity,
                        role=permission_type.description.lower()),
                    category=u"info")
            elif form.action.data == 'delete':
                # TRANS: Message shown when trying to remove a permission from
                # TRANS: a user, but that user didn't have that permission
                # TRANS: already.
                flash(gettext(u"%(name)s is not a %(role)s.",
                        name=entity,
                        role=permission_type.description.lower()),
                    category=u"warning")
        else:
            if form.action.data == 'delete':
                permission_query.delete()
                # TRANS: Confirmation message shown when revoking a permission
                # TRANS: from a user or group.
                flash(gettext(u"%(name)s is no longer a %(role)s.",
                        name=entity,
                        role=permission_type.description.lower()),
                    category=u"info")
            elif form.action.data == 'add':
                flash(gettext(u"%(name)s is now a %(role)s.",
                        name=entity,
                        role=permission_type.description.lower()),
                    category=u"info")
        db.session.commit()
    else:
        for field_name, errors in six.iteritems(form.errors):
            errors = u", ".join(errors)
            # TRANS: Error message that is shown when one or more fields of a
            # TRANS: form are shown.
            flash(gettext(u"Errors for %(field_name)s: %(error)s.",
                field_name=field_name, error=errors), u'error')
        current_app.logger.info("Malformed entity permission POST: {}".format(
                form.errors))
    return get_division_details(division=division)


def _modify_division_transformer(division):
    """Handle POST requests for changing the Transformers for a Division."""
    form = ChangeTransformer()
    # Set the form's choices
    attr = form.attribute.data
    form.transformer.choices = transformer_choices(attr)
    # Check form and go from there
    if form.validate():
        name = form.transformer.data
        if name == 'none':
            division.transformers[attr] = None
        else:
            # Get the specific map of transformers for the attribute
            attr_transformers = current_app.url_transformers[attr]
            # Get the new transformer
            division.transformers[attr] = attr_transformers[name]
            # Explicitly add the TransformerRef to the session
            db.session.add(division.division_transformers[attr])
        db.session.commit()
        # TRANS: Confirmation message shown when a transformer for an
        # TRANS: attribute has been set.
        flash(gettext(u"'%(attribute)s' set to '%(transformer)s'.",
                attribute=attr, transformer=name), u'message')
    else:
        for field_name, errors in six.iteritems(form.errors):
            errors = u", ".join(errors)
            # TRANS: Generic error message shown for the fields in a form.
            flash(gettext(u"Errors for %(field_name)s: %(error)s.",
                field_name=field_name, error=errors), u'error')
        current_app.logger.info("Malformed division transformer POST: {}".
                format(form.errors))
    return get_division_details(division=division)


@blueprint.route('/<int:division_id>/', methods=['POST'])
@fresh_login_required
def modify_division(division_id):
    """Dispatches modification requests to the specialized view function for
    that operation.
    """
    division = Division.query.get_or_404(division_id)
    if not current_user.admin and not \
            current_user.has_permission(PermissionType.admin, division):
        abort(403)
    form_id = request.form.get('form_id')
    if form_id == 'entity':
        return _modify_division_entity(division)
    elif form_id == 'transformer':
        return _modify_division_transformer(division)
    else:
        current_app.logger.warn("Invalid division modification POST: {}"
                .format(request.form))
        abort(400)


@blueprint.route('/<int:division_id>/transformers/')
@blueprint.route('/<int:division_id>/transformers/<attribute>/')
@login_required
def list_transformers(division_id, attribute=None):
    """API method to get a list of transformers for a division.

    :param division_id int: the ID of the division to look up
    :param attribute str: a specific attribute to look up. Optional.
    :return: JSON
    """
    division = Division.query.get_or_404(division_id)
    if not current_user.admin and not \
            current_user.has_permission(PermissionType.admin, division):
        abort(403)
    if attribute is None:
        attrs = six.iterkeys(current_app.url_transformers)
    else:
        attrs = (attribute,)
    choices = {}
    for attr in attrs:
        raw_choices = transformer_choices(attr)
        current = division.transformers.get(attr, None)
        if current is not None:
            choices[attr] = \
                    [(c[0], c[1], c[1] == current.name) for c in raw_choices]
        else:
            choices[attr] = \
                    [(c[0], c[1], False) for c in raw_choices]
    return jsonify(choices)



from __future__ import absolute_import
from decimal import Decimal
import locale
import os
import requests
import sys
import warnings
from flask import current_app, g
import flask_sqlalchemy
from flask_babel import Babel, get_locale
from flask_wtf.csrf import CsrfProtect
import six
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import MetaData
from werkzeug.utils import import_string
from .transformers import Transformer
from .versioned_static import static_file, VersionedStaticFlask

try:
    import raven.contrib.flask as raven_flask
    sentry = raven_flask.Sentry()
except ImportError:
    sentry = None


db = flask_sqlalchemy.SQLAlchemy()
# Patch Flask-SQLAlchemy to use a custom Metadata instance with a naming scheme
# for constraints.
def _patch_metadata():
    naming_convention = {
        'fk': ('fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s'
                '_%(referred_column_0_name)s'),
        'pk': 'pk_%(table_name)s',
        'ix': 'ix_%(table_name)s_%(column_0_name)s',
        'ck': 'ck_%(table_name)s_%(constraint_name)s',
        'uq': 'uq_%(table_name)s_%(column_0_name)s',
    }
    metadata = MetaData(naming_convention=naming_convention)
    base = declarative_base(cls=flask_sqlalchemy.Model, name='Model',
                            metaclass=flask_sqlalchemy._BoundDeclarativeMeta,
                            metadata=metadata)
    base.query = flask_sqlalchemy._QueryProperty(db)
    db.Model = base
_patch_metadata()

# Work around buggy HTTP servers sending out 0 length chunked responses
def _patch_httplib():
    import six.moves.http_client
    def patch_http_response_read(func):
        def inner(*args):
            try:
                return func(*args)
            except six.moves.http_client.IncompleteRead as e:
                return e.partial

        return inner
    six.moves.http_client.HTTPResponse.read = patch_http_response_read(
            six.moves.http_client.HTTPResponse.read)
_patch_httplib()

from .util import DB_STATS, AcceptRequest, WeakCiphersAdapter


__version__ = u'0.12.11'


requests_session = requests.Session()


# Set default locale
locale.setlocale(locale.LC_ALL, '')


csrf = CsrfProtect()


babel = Babel()


# Ensure models are declared
from . import models
from .auth import models as auth_models
from .auth import AuthMethod


def create_app(config=None, **kwargs):
    """Create the WSGI application that is this app.

    In addition to the arguments documented below, this function accepts as a
    keyword agument all agruments to :py:class:`flask.Flask` except
    `import_name`, which is set to 'evesrp'. Additionally, the value of
    `instance_relative_config` has a default value of `True`.

    If the `config` argument isn't specified, the app will attempt to use the
    file 'config.py' in the instance folder if it exists, and will then fall
    back to using the value of the EVESRP_SETTINGS environment variable as a
    path to a config file.

    :param config: The app configuration file. Can be a Python
        :py:class:`dict`, a path to a configuration file, or an importable
        module name containing the configuration.
    :type config: str, dict
    """
    # Default instance_relative_config to True to let the config fallback work
    kwargs.setdefault('instance_relative_config', True)
    app = VersionedStaticFlask('evesrp', **kwargs)
    app.request_class = AcceptRequest
    app.config.from_object('evesrp.default_config')
    # Push the instance folder path onto sys.path to allow importing from there
    sys.path.insert(0, app.instance_path)
    # Check in config is a dict, python config file, or importable object name,
    # in that order. Finally, check the EVESRP_SETTINGS environment variable
    # as a last resort.
    if isinstance(config, dict):
        app.config.update(config)
    elif isinstance(config, six.string_types):
        if config.endswith(('.txt', '.py', '.cfg')):
            app.config.from_pyfile(config)
        else:
            app.config.from_object(config)
    elif config is None:
        try:
            app.config.from_pyfile('config.py')
        except OSError:
            app.config.from_envvar('EVESRP_SETTINGS')

    # Configure Sentry
    if 'SENTRY_DSN' in app.config or 'SENTRY_DSN' in os.environ:
        if sentry is not None:
            app.config['SENTRY_RELEASE'] = __version__
            sentry.init_app(app=app)
        else:
            app.logger.warning("SENTRY_DSN is defined but Sentry is not"
                               " installed.")

    # Register SQLAlchemy monitoring before the DB is connected
    app.before_request(sqlalchemy_before)

    db.init_app(app)

    from .views.login import login_manager
    login_manager.init_app(app)

    before_csrf = list(app.before_request_funcs[None])
    csrf.init_app(app)
    # Remove the context processor that checks CSRF values. All it is used for
    # is the template function.
    app.before_request_funcs[None] = before_csrf

    # Connect views
    from .views import index, error_page, update_navbar, divisions, login,\
            requests, api, detect_language, locale_selector
    app.add_url_rule(rule=u'/', view_func=index)
    for error_code in (400, 403, 404, 500):
        app.register_error_handler(error_code, error_page)
    app.after_request(update_navbar)
    app.register_blueprint(divisions.blueprint, url_prefix='/division')
    app.register_blueprint(login.blueprint)
    app.register_blueprint(requests.blueprint, url_prefix='/request')
    app.register_blueprint(api.api, url_prefix='/api')
    app.register_blueprint(api.filters, url_prefix='/api/filter')

    from .views import request_count
    app.add_template_global(request_count)

    from .json import SRPEncoder
    app.json_encoder=SRPEncoder

    # Hook up Babel and associated callbacks
    babel.init_app(app)
    app.before_request(detect_language)
    # localeselector can be set only once per Babel instance. Really, this will
    # only throw an exception when we're running tests. This also only became
    # an issue when they changed when Babel.locale_selector_func was changed
    # for version 0.10.
    try:
        babel.localeselector(locale_selector)
    except AssertionError:
        pass

    # Configure the Jinja context
    # Inject variables into the context
    from .auth import PermissionType
    from .util import locale as jinja_locale
    @app.context_processor
    def inject_enums():
        return {
            'ActionType': models.ActionType,
            'PermissionType': PermissionType,
            'app_version': __version__,
            'site_name': app.config['SRP_SITE_NAME'],
            'url_for_page': requests.url_for_page,
            'static_file': static_file,
            'locales': jinja_locale.enabled_locales,
            'get_locale': get_locale,
        }
    app.template_filter('currencyfmt')(jinja_locale.currencyfmt)
    app.template_filter('percentfmt')(jinja_locale.percentfmt)
    app.template_filter('numberfmt')(jinja_locale.numberfmt)
    # Auto-trim whitespace
    app.jinja_env.trim_blocks = True
    app.jinja_env.lstrip_blocks = True

    init_app(app)

    return app


def init_app(app):
    _config_requests_session(app)
    _config_url_converters(app)
    _config_authmethods(app)
    _config_killmails(app)


# SQLAlchemy performance logging
def sqlalchemy_before():
    if DB_STATS.total_queries > 0:
        current_app.logger.debug(u"{} queries in {} ms.".format(
                DB_STATS.total_queries,
                round(DB_STATS.total_time * 1000, 3)))
    DB_STATS.clear()
    g.DB_STATS = DB_STATS


# Utility function for creating instances from dicts
def _instance_from_dict(instance_descriptor):
    type_name = instance_descriptor.pop('type')
    Type = import_string(type_name)
    return Type(**instance_descriptor)


# Utility function for raising config deprecation warnings
def _deprecated_object_instance(key, value):
    warnings.warn(u"Non-basic data types in configuration values are deprecated"
                 u"({}: {})".format(key, value), DeprecationWarning,
                 stacklevel=2)


# Auth setup
def _config_authmethods(app):
    auth_methods = []
    # Once the deprecated config value support is removed, this can be
    # rewritten as a dict comprehension
    with app.app_context():
        for method in app.config['SRP_AUTH_METHODS']:
            if isinstance(method, dict):
                auth_methods.append(_instance_from_dict(method))
            elif isinstance(method, AuthMethod):
                _deprecated_object_instance('SRP_AUTH_METHODS', method)
                auth_methods.append(method)
    app.auth_methods = auth_methods


# Request detail URL setup
def _config_url_converters(app):
    url_transformers = {}
    for config_key, config_value in app.config.items():
        # Skip null config values
        if config_value is None:
            continue
        # Look for config keys in the format 'SRP_*_URL_TRANSFORMERS'
        if not config_key.endswith('_URL_TRANSFORMERS')\
                or not config_key.startswith('SRP_'):
            continue
        attribute = config_key.replace('SRP_', '', 1)
        attribute = attribute.replace('_URL_TRANSFORMERS', '', 1)
        attribute = attribute.lower()
        # Create Transformer instances for this attribute
        transformers = {}
        for transformer_config in config_value:
            if isinstance(transformer_config, tuple):
                # Special case for Transformers: A 2-tuple in the form:
                # (u'Transformer Name',
                #     'http://example.com/transformer/slug/{}')
                transformer = Transformer(*transformer_config)
            elif isinstance(transformer_config, dict):
                # Standard instance dictionary format
                # Provide a default type value
                transformer_config.setdefault('type',
                        'evesrp.transformers.Transformer')
                transformer = _instance_from_dict(transformer_config)
            elif isinstance(transformer_config, Transformer):
                # DEPRECATED: raw Transformer instance
                _deprecated_object_instance(config_key, transformer_config)
                transformer = transformer_config
            transformers[transformer.name] = transformer
        url_transformers[attribute] = transformers
    app.url_transformers = url_transformers


# Requests session setup
def _config_requests_session(app):
    try:
        ua_string = app.config['SRP_USER_AGENT_STRING']
    except KeyError as outer_exc:
        try:
            ua_string = 'EVE-SRP/{version} ({email})'.format(
                    email=app.config['SRP_USER_AGENT_EMAIL'],
                    version=__version__)
        except KeyError as inner_exc:
            raise inner_exc
    requests_session = requests.Session()
    requests_session.headers.update({'User-Agent': ua_string})
    requests_session.mount('https://crest-tq.eveonline.com',
            WeakCiphersAdapter())
    app.requests_session = requests_session


# Killmail verification
def _config_killmails(app):
    killmail_sources = []
    # For now, use a loop with checks. After removing the depecated config
    # method it can be rewritten as a list comprehension
    for source in app.config['SRP_KILLMAIL_SOURCES']:
        if isinstance(source, six.string_types):
            killmail_sources.append(import_string(source))
        elif isinstance(source, type):
            _deprecated_object_instance('SRP_KILLMAIL_SOURCES', source)
            killmail_sources.append(source)
    app.killmail_sources = killmail_sources


# Work around DBAPI-specific issues with Decimal subclasses.
# Specifically, almost everything besides pysqlite and psycopg2 raise
# exceptions if an instance of a Decimal subclass as opposed to an instance of
# Decimal itself is passed in as a value for a Numeric column.
@db.event.listens_for(Engine, 'before_execute', retval=True)
def _workaround_decimal_subclasses(conn, clauseelement, multiparams, params):
    def filter_decimal_subclasses(parameters):
        for key in six.iterkeys(parameters):
            value = parameters[key]
            # Only get instances of subclasses of Decimal, not direct
            # Decimal instances
            if type(value) != Decimal and isinstance(value, Decimal):
                parameters[key] = Decimal(value)

    if multiparams:
        for mp in multiparams:
            if isinstance(mp, dict):
                filter_decimal_subclasses(mp)
            elif isinstance(mp, list):
                for parameters in mp:
                    filter_decimal_subclasses(parameters)
    return clauseelement, multiparams, params



from __future__ import absolute_import
import re

from flask import redirect, url_for, current_app
from flask_babel import gettext, lazy_gettext
import flask_login
from flask_wtf import Form
from wtforms.fields import SubmitField, HiddenField
from ..util import DeclEnum, classproperty, ensure_unicode


class AuthMethod(object):
    """Represents an authentication mechanism for users."""

    def __init__(self, admins=None, name=u'Base Authentication', **kwargs):
        """
        :param admins: A list of usernames to treat as site-wide
            administrators. Useful for initial setup.
        :type admins: list
        :param str name: The user-facing name for this authentication method.
        """
        if admins is None:
            self.admins = []
        else:
            self.admins = admins
        self.name = ensure_unicode(name)

    def form(self):
        """Return a :py:class:`flask_wtf.Form` subclass to login with."""
        class AuthForm(Form):
            submit = SubmitField(gettext(
                    u'Log In using %(authmethod_name)s',
                    authmethod_name=self.name))
        return AuthForm

    def login(self, form):
        """Process a validated login form.

        You must return a valid response object.
        """
        pass

    def view(self):
        """Optional method for providing secondary views.

        :py:func:`evesrp.views.login.auth_method_login` is configured to allow
        both GET and POST requests, and will call this method as soon as it is
        known which auth method is meant to be called. The path for this view
        is ``/login/self.safe_name/``, and can be generated with
        ``url_for('login.auth_method_login', auth_method=self.safe_name)``.

        The default implementation redirects to the main login view.
        """
        return redirect(url_for('login.login'))

    @staticmethod
    def login_user(user):
        """Signal to the authentication systems that a new user has logged in.

        Handles calling :py:func:`flask_login.login_user` and any other
        related housekeeping functions for you.

        :param user: The user that has been authenticated and is logging in.
        :type user: :py:class:`~models.User`
        """
        flask_login.login_user(user)

    @property
    def safe_name(self):
        """Normalizes a string to be a valid Python identifier (along with a few
        other things).

        Specifically, all letters are lower cased and non-ASCII and whitespace
        are replaced by underscores.

        :returns: The normalized string.
        :rtype str:
        """
        # Turn 'fancy' characters into '?'s
        ascii_rep = self.name.encode('ascii', 'replace').decode('utf-8')
        # Whitespace and '?' to underscores
        no_space = re.sub(r'[\s\?]', u'_', ascii_rep)
        lowered = no_space.lower()
        return lowered

    def refresh(self, user):
        """Refresh a user's information (if possible).

        The :py:class:`~.AuthMethod` should attmept to refresh the given user's
        information as if they were logging in for the first time.

        :param user: The user to refresh.
        :type user: :py:class:`~models.User`
        :returns: Wether or not the refresh attempt succeeded.
        :rtype: :py:class:`bool`
        """
        return False


class AnonymousUser(flask_login.AnonymousUserMixin):

    def has_permission(self, permissions, division_or_request=None):
        return False


class PermissionType(DeclEnum):
    """Enumerated type for the types of permissions available. """

    # TRANS: The title for someone who is able to submit a request to a
    # TRANS: division.
    submit = u'submit', lazy_gettext(u'Submitter')
    """Permission allowing submission of :py:class:`~.Request`\s to a
    :py:class:`~.Division`.
    """

    # TRANS: The title for someone who is able to review requests to a
    # TRANS: division.
    review = u'review', lazy_gettext(u'Reviewer')
    """Permission for reviewers of requests in a :py:class:`~.Division`."""

    # TRANS: The title form someone who is able to mark requests as paid in a
    # TRANS: division.
    pay = u'pay', lazy_gettext(u'Payer')
    """Permission for payers in a :py:class:`~.Division`."""

    # TRANS: The title for someone who is able to administer a division.
    admin = u'admin', lazy_gettext(u'Administrator')
    """:py:class:`~.Division`\-level administrator permission"""

    # TRANS: The title for someone who is able to audit the activities of a
    # TRANS: division. This means they are able to view all requests in a
    # TRANS: division.
    audit = u'audit', lazy_gettext(u'Auditor')
    """A special permission for allowing read-only elevated access"""

    @classproperty
    def elevated(cls):
        return frozenset((cls.review, cls.pay, cls.admin, cls.audit))

    @classproperty
    def all(cls):
        return frozenset((cls.submit,
                          cls.review,
                          cls.pay,
                          cls.admin,
                          cls.audit))



from __future__ import absolute_import
from __future__ import unicode_literals
import datetime as dt
from sqlalchemy.types import DateTime
from sqlalchemy.ext.declarative import declared_attr
from .datetime import DateTime
from .datetime import utc
from .. import db


class AutoID(object):
    """Mixin adding a primary key integer column named 'id'."""
    id = db.Column(db.Integer, primary_key=True)

    def _json(self, extended=False):
        try:
            parent = super(AutoID, self)._json(extended)
        except AttributeError:
            parent = {}
        parent[u'id'] = self.id
        return parent


def _utcnow(arg):
    return dt.datetime.now(utc)

class Timestamped(object):
    """Mixin adding a timestamp column.

    The timestamp defaults to the current time.
    """
    timestamp = db.Column(DateTime, nullable=False,
            default=_utcnow)


class AutoName(object):

    @declared_attr
    def __tablename__(cls):
        """SQLAlchemy late-binding attribute to set the table name.

        Implemented this way so subclasses do not need to specify a table name
        themselves.
        """
        return cls.__name__.lower()



from flask import flash, current_app


def check_crest_response(response):
    """Check for CREST representation deprecation/removal.

    Specifically, check that the status code isn't 406 (meaning the
    representation has been removed) and for the presence of the X-Deprecated
    header.

    :param Response response: A :py:class:`~.Response` to check.
    :rtype: bool
    """
    # TODO Add a test case for this
    if response.status_code == 406:
        flash((u"This version of EVE SRP no longer knows how to interface with"
               u"the CREST API. Please update to the latest version."),
               u'error')
        return False
    if 'X-Deprecated' in response.headers:
        flash((u"The version of the CREST representation known by EVE SRP "
               u"has been deprecated. Please update to the latest version "
               u"to ensure continued operation."), u'warn')
    return True


class NameLookup(object):

    def __init__(self, starting_data, root_key, content_type, attribute='name'):
        self._dict = starting_data
        self.content_type = content_type
        self.attribute_name = attribute
        self.root_key = root_key

    @property
    def url_slug(self):
        if not hasattr(self, '_url_slug'):
            root_resp = current_app.requests_session.get(
                    'https://crest-tq.eveonline.com/')
            root_json = root_resp.json()
            url_slug = root_json[self.root_key]['href']
            url_slug += '{}/'
            self._url_slug = url_slug
        return self._url_slug

    def _access_attribute(self, obj, attribute=None):
        if attribute is None:
            attribute = self.attribute_name
        if '.' in attribute:
            first, rest = self.attribute_name.split('.', 1)
            return self._access_attribute(obj[first], rest)
        else:
            return obj[attribute]

    def __getitem__(self, key):
        if not isinstance(key, int):
            raise TypeError("Invalid ID for {} lookup: '{}'".\
                    format(self.attribute_name, key))
        if key not in self._dict:
            resp = current_app.requests_session.get(
                    self.url_slug.format(key),
                    headers={'Accept': self.content_type})
            if check_crest_response(resp) and resp.status_code == 200:
                self._dict[key] = self._access_attribute(resp.json())
            else:
                message = "Cannot find the {} for the ID requested [{}]: {}"\
                        .format(self.attribute_name, resp.status_code, key)
                raise KeyError(message)
        return self._dict[key]



"""Allow Pilots to not have owning users.

Revision ID: e6b2c156b11
Revises: 8f179e969c9
Create Date: 2014-09-03 17:13:17.282827

"""

# revision identifiers, used by Alembic.
revision = 'e6b2c156b11'
down_revision = '8f179e969c9'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

def upgrade():
    op.alter_column('pilot', 'user_id', existing_type=sa.Integer,
            nullable=True)


def downgrade():
    op.alter_column('pilot', 'user_id', existing_type=sa.Integer,
            nullable=False)



from __future__ import absolute_import
import datetime as dt
from flask import flash, redirect, current_app, url_for, request, session
from flask_babel import gettext
from flask_login import current_user, login_user, login_fresh
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import OAuth2Error
from flask_wtf import csrf
import six
from sqlalchemy.orm.exc import NoResultFound

from .. import db, sentry
from ..util import ensure_unicode, DateTime, is_safe_redirect
from . import AuthMethod
from .models import User, Group, Pilot


class OAuthMethod(AuthMethod):

    def __init__(self, **kwargs):
        """Abstract :py:class:`~.AuthMethod` for OAuth-based login methods.

        Implementing classes need to implement :py:meth:`get_user`\,
        :py:meth:`get_pilots`\, and :py:meth:`get_groups`\.

        In addition to the keyword arguments from :py:class:`~.AuthMethod`,
        this initializer accepts the following arguments that will be used in
        the creation of the :py:attr:`OAuthMethod.oauth` object (See the
        documentation for :py:class:`~flask_oauthlib.client.OAuthRemoteApp` for
        more details):

        * ``client_id``
        * ``client_secret``
        * ``scope``
        * ``access_token_url``
        * ``refresh_token_url``
        * ``authorize_url``
        * ``access_token_params``
        * ``method``

        As a convenience, the ``key`` and ``secret`` keyword arguments will be
        treated as ``consumer_key`` and ``consumer_secret`` respectively. The
        ``name`` argument is used for both :py:class:`~.AuthMethod` and for
        :py:class:`~flask_oauthlib.client.OAuthRemoteApp`.

        Subclasses for providers that may be used by more than one entity are
        encouraged to provide their own defaults for the above arguments.

        The redirect URL for derived classes is based off of the
        :py:attr:`~.AuthMethod.safe_name` of the implementing
        :py:class:`~.AuthMethod`, specifically the URL for
        :py:meth:`~.AuthMethod.view`. For example, the default redirect URL for
        :py:class:`~.TestOAuth` is similar to
        ``https://example.com/login/test_oauth/`` (Note the trailing slash, it
        is significant).

        :param default_token_expiry: The default time (in seconds) access
            tokens are valid for. Defaults to 5 minutes.
        :type default_token_expiry: :py:class:`int`
        """
        keyword_mapping = {
            'key': 'client_id',
            'consumer_key': 'client_id',
            'secret': 'client_secret',
            'consumer_secret': 'client_secret',
            'access_token_method': 'method',
        }
        for old_kw, new_kw in six.iteritems(keyword_mapping):
            if old_kw in kwargs:
                # TODO: Raise a deprecation warning
                kwargs[new_kw] = kwargs.pop(old_kw)
        # TODO: add handling for old request_token_params and pulling scope out?
        # Save arguments used for later operations
        self.client_id = kwargs.pop('client_id')
        self.client_secret = kwargs.pop('client_secret')
        self.authorize_url = kwargs.pop('authorize_url')
        self.token_url = kwargs.pop('access_token_url')
        self.refresh_url = kwargs.pop('refresh_token_url', None)
        self.oauth_method = kwargs.pop('method', 'POST')
        self.scope = kwargs.pop('scope', None)
        self.default_token_expiry = kwargs.pop('default_token_expiry', 300)
        self.name = kwargs.get('name', 'OAuth')
        try:
            self.access_params = kwargs.pop('access_token_params')
        except KeyError:
            self.access_params = {}

        # Small function that will mark users as stale (not fresh) for
        # Flask-Login if they no longer have valid tokens. This will trigger a
        # refresh if there are refresh tokens, or require them to login on the
        # next attempted privileged action.
        @current_app.before_request
        def enforce_user_freshness():
            if current_user.is_authenticated and \
                    current_user.authmethod == self.name and \
                    login_fresh():
                user = current_user._get_current_object()
                if user.seconds_valid <= 0:
                    current_app.logger.debug("Marking '{}' as stale".format(
                        user))
                    login_user(user, fresh=False)
                    self.session.token = user.token

        super(OAuthMethod, self).__init__(**kwargs)

    # Being done in a property so when url_for is called, it has access to a
    # request, specifically the scheme.
    @property
    def redirect_uri(self):
        return  url_for('login.auth_method_login',
                auth_method=self.safe_name, _external=True)


    def login(self, form):
        oauth = OAuth2Session(self.client_id,
                redirect_uri=self.redirect_uri,
                scope=self.scope)
        url, state = oauth.authorization_url(self.authorize_url)
        session['state'] = state
        current_app.logger.debug(u"Redirecting to : {}".format(url))
        # Stash the 'next' parameter in the session to use it in the 'view'
        # view. It's automatically added by Flask-Login.
        if 'next' in request.args:
            session['next'] = request.args['next']
        return redirect(url)

    def view(self):
        """Handle creating and/or logging in the user and updating their
        :py:class:`~.Pilot`\s and :py:class:`~.Group`\s.
        """
        oauth = OAuth2Session(self.client_id,
                redirect_uri=self.redirect_uri,
                state=session['state'])
        try:
            token = oauth.fetch_token(self.token_url,
                    authorization_response=request.url,
                    method=self.oauth_method,
                    client_secret=self.client_secret,
                    auth=(self.client_id, self.client_secret))
        except OAuth2Error as e:
            # TRANS: When there's an error associated with a login.
            flash(gettext(u"Login failed: %(error)s", error=e.error))
            if 'SENTRY_RELEASE' in current_app.config:
                sentry.captureException()
            return redirect(url_for('login.login'))
        # Set the current session manually because the automated method relies
        # on current_user.
        self.session = OAuth2Session(self.client_id, token=token)
        # Get the User object for this user, creating one if needed
        user = self.get_user()
        # Update the tokens (and related info) for the user
        user.access_token = token[u'access_token']
        # If not given by the server, use the default expiry time
        if u'expires_in' in token:
            user.set_expiration(token[u'expires_in'])
        else:
            user.set_expiration(self.default_token_expiry)
        user.refresh_token = token.get(u'refresh_token')
        db.session.commit()
        if user is not None:
            # Login the user, so current_user will work
            self.login_user(user)
        else:
            # TRANS: Error shown for a failed login.
            flash(gettext(u"Login failed."), u'error')
            return redirect(url_for('login.login'))
        # Update admins, groups and pilots for the current user
        self._update_user_info()
        # Redirect to the 'next' parameter given to the 'login' view.
        # The next parameter is automatically added by Flask-Login.
        # Check that the 'next' parameter is safe.
        next_url = request.args.get('next')
        if next_url is not None:
            if not is_safe_redirect(next_url):
                next_url = None
        return redirect(next_url or url_for('index'))


    def _update_user_info(self):
        current_app.logger.debug(
                "Updating information for '{}' with OAuth".format(current_user))
        # Set the site-wide admin flag
        current_user.admin = self.is_admin(current_user)
        # Add new Pilots
        current_pilots = self.get_pilots()
        for pilot in current_pilots:
            pilot.user = current_user
        # Remove old pilots
        user_pilots = set(current_user.pilots)
        for pilot in user_pilots:
            if pilot not in current_pilots:
                pilot.user = None
        # Add new groups
        current_groups = self.get_groups()
        for group in current_groups:
            current_user.groups.add(group)
        # Remove old groups
        user_groups = set(current_user.groups)
        for group in user_groups:
            if group not in current_groups and group in current_user.groups:
                current_user.groups.remove(group)
        # Save all changes
        db.session.commit()

    def get_user(self):
        """Returns the :py:class:`~.OAuthUser` instance for the current token.

        This method is to be implemented by subclasses of
        :py:class:`OAuthMethod` to use whatever APIs they have access to to get
        the user account given an access token.

        :rtype: :py:class:`OAuthUser`
        """
        raise NotImplementedError

    def is_admin(self, user):
        """Returns wether this user should be treated as a site-wide
        administrator.

        The default implementation checks if the user's name is contained
        within the list of administrators supplied as an argument to
        :py:class:`OAuthMethod`\.

        :param user: The user to check.
        :type user: :py:class:`~.OAuthUser`
        :rtype: bool
        """
        return user.name in self.admins

    def get_pilots(self):
        """Return a :py:class:`list` of :py:class:`~.Pilot`\s for the given
        token.

        Like :py:meth:`get_user`\, this method is to be implemented by
        :py:class:`OAuthMethod` subclasses to return a list of
        :py:class:`~.Pilot`\s associated with the account for the given access
        token.

        :rtype: :py:class:`list` of :py:class:`~.Pilot`\s.
        """
        raise NotImplementedError

    def get_groups(self):
        """Returns a :py:class:`list` of :py:class:`~.Group`\s for the given
        token.

        Like :py:meth:`get_user` and :py:meth:`get_pilots`\, this method is to
        be implemented by :py:class:`OAuthMethod` subclasses to return a list
        of :py:class:`~.Group`\s associated with the account for the given
        access token.

        :rtype: :py:class:`list` of :py:class:`~.Group`\s.
        """
        raise NotImplementedError

    def refresh(self, user):
        """Refreshes the current user's information.

        Attempts to refresh the pilots and groups for the given user. If the
        current access token has expired, the refresh token is used to get a
        new access token.
        """
        try:
            self._update_user_info()
        except OAuth2Error:
            return False
        else:
            return True

    @property
    def session(self):
        if not hasattr(self, '_oauth_session'):
            if not current_user.is_anonymous:
                kwargs = {
                    'token': current_user.token,
                }
                if self.refresh_url is not None:
                    kwargs['auto_refresh_url'] = self.refresh_url
                    kwargs['token_updater'] = token_saver
                    kwargs['auto_refresh_kwargs'] = {
                        'client_id': self.client_id,
                        'client_secret': self.client_secret,
                    }
                current_app.logger.debug(u"Creating a new OAuth2Session for "
                                         u"'{}'".format(current_user))
                self._oauth_session = OAuth2Session(self.client_id, **kwargs)
        return self._oauth_session

    @session.setter
    def session(self, new_session):
        self._oauth_session = new_session

    @session.deleter
    def session(self):
        del self._oauth_session


def token_saver(token):
    current_app.logger.debug("Saving new OAuth tokens for user '{}'".format(
        current_user))
    current_user.access_token = token[u'access_token']
    refresh_token = token.get(u'refresh_token')
    if refresh_token is not None and \
            refresh_token == current_user.refresh_token:
        current_app.logger.warning(u"Refresh token did not change.")
    else:
        current_user.refresh_token = refresh_token
    current_user.set_expiration(token.get(u'expires_in', 0))
    db.session.commit()


class OAuthUser(User):

    id = db.Column(db.Integer, db.ForeignKey(User.id), primary_key=True)

    access_token = db.Column(db.String(100, convert_unicode=True))

    access_expiration = db.Column(DateTime, default=None)

    refresh_token = db.Column(db.String(100, convert_unicode=True))

    def set_expiration(self, expiration_seconds):
        now = dt.datetime.utcnow()
        expiration = now + dt.timedelta(seconds=expiration_seconds)
        self.access_expiration = expiration

    @property
    def seconds_valid(self):
        if self.access_expiration is not None:
            now = dt.datetime.utcnow()
            current_duration = self.access_expiration - now
            return current_duration.total_seconds()
        else:
            return 0

    @property
    def token(self):
        token = {'token_type': 'Bearer'}
        token['access_token'] = self.access_token
        token['expires_in'] = int(self.seconds_valid)
        if self.refresh_token is not None:
            token['refresh_token'] = self.refresh_token
        return token



from .classproperty import classproperty
from .decimal import PrettyDecimal, PrettyNumeric
from .enum import DeclEnum
from .jsonify import jsonify
from .models import AutoID, Timestamped, AutoName
from .request import AcceptRequest
from .sqlstats import DB_STATS
from .unistr import unistr, ensure_unicode
from .urlparse import urlparse, urlunparse
from .datetime import utc, DateTime, parse_datetime
from .varies import varies
from .validate_redirect import is_safe_redirect
from .weak_ciphers import WeakCiphersAdapter
from .xmlify import xmlify



class Transformer(object):

    def __init__(self, name, slug):
        self.name = name
        self.slug = slug

    def __repr__(self):
        return u"{x.__class__.__name__}('{x.name}', '{x.slug}')".format(x=self)

    def __str__(self):
        return self.name

    def __call__(self, *args, **kwargs):
        return self.slug.format(*args, **kwargs)

    def __hash__(self):
        return hash(self.name + self.slug) ^ hash(self.__class__)

    def __eq__(self, other):
        return hash(self) == hash(other)



import six
from six.moves.urllib import parse

class UnicodeParseResult(parse.ParseResult):

    @staticmethod
    def ensure_unicode(value):
        if isinstance(value, six.binary_type):
            value = value.decode()
        return value

    def __getattr__(self, attr):
        value = super(UnicodeParseResult, self).__getattr__(attr)
        return self.ensure_unicode(value)

    def __getitem__(self, key):
        value = super(UnicodeParseResult, self).__getitem__(key)
        return self.ensure_unicode(value)


def urlparse(*args, **kwargs):
    real_parse = parse.urlparse(*args, **kwargs)
    return UnicodeParseResult(*real_parse)

urlunparse = parse.urlunparse



from __future__ import absolute_import
from collections import defaultdict
import datetime as dt
import time
from decimal import Decimal
from functools import partial
import re
import sys
import six
from .util import unistr, urlparse, urlunparse, utc, classproperty

from flask import Markup, current_app
from flask_babel import gettext, lazy_gettext
import iso8601
import requests
from sqlalchemy import create_engine, Table, MetaData
from sqlalchemy.sql import select

from . import ships, systems


if six.PY3:
    unicode = str


@unistr
class Killmail(object):
    """Base killmail representation.

    .. py:attribute:: kill_id

        The ID integer of this killmail. Used by most killboards and by CCP to
        refer to killmails.

    .. py:attribute:: ship_id

        The typeID integer of for the ship lost for this killmail.

    .. py:attribute:: ship

        The human readable name of the ship lost for this killmail.

    .. py:attribute:: pilot_id

        The ID number of the pilot who lost the ship. Referred to by CCP as
        ``characterID``.

    .. py:attribute:: pilot

        The name of the pilot who lost the ship.

    .. py:attribute:: corp_id

        The ID number of the corporation :py:attr:`pilot` belonged to at the
        time this kill happened.

    .. py:attribute:: corp

        The name of the corporation referred to by :py:attr:`corp_id`.

    .. py:attribute:: alliance_id

        The ID number of the alliance :py:attr:`corp` belonged to at the time
        of this kill, or ``None`` if the corporation wasn't in an alliance at
        the time.

    .. py:attribute:: alliance

        The name of the alliance referred to by :py:attr:`alliance_id`.

    .. py:attribute:: url

        A URL for viewing this killmail's information later. Typically an
        online killboard such as `zKillboard <https://zkillboard.com>`_, but
        other kinds of links may be used.

    .. py:attribute:: value

        The extimated ISK loss for the ship destroyed in this killmail. This is
        an optional attribute, and is ``None`` if unsupported. If this
        attribute is set, it should be a :py:class:`~.Decimal` or at least a
        type that can be used as the value for the :py:class:`~.Decimal`
        constructor.

    .. py:attribute:: timestamp

        The date and time that this kill occured as a
        :py:class:`datetime.datetime` object (with a UTC timezone).

    .. py:attribute:: verified

        Whether or not this killmail has been API verified (or more accurately,
        if it is to be trusted when making a
        :py:class:`~evesrp.models.Request`.

    .. py:attribute:: system

        The name of the system where the kill occured.

    .. py:attribute:: system_id

        The ID of the system where the kill occured.

    .. py:attribute:: constellation

        The name of the constellation where the kill occured.

    .. py:attribute:: region

        The name of the region where the kill occured.
    """

    def __init__(self, **kwargs):
        """Initialize a :py:class:`Killmail` with ``None`` for all attributes.

        All subclasses of this class, (and all mixins designed to be used with
        it) must call ``super().__init__(**kwargs)`` to ensure all
        initialization is done.

        :param: keyword arguments corresponding to attributes.
        """
        self._data = defaultdict(lambda: None)
        for attr in ('kill_id', 'ship_id', 'ship', 'pilot_id', 'pilot',
                     'corp_id', 'corp', 'alliance_id', 'alliance', 'verified',
                     'url', 'value', 'timestamp', 'system', 'constellation',
                     'region', 'system_id'):
            try:
                setattr(self, attr, kwargs[attr])
            except KeyError:
                pass
            else:
                del kwargs[attr]
        super(Killmail, self).__init__(**kwargs)

    # Any attribute not starting with an underscore will be stored in a
    # separate, private attribute. This is to allow attributes on Killmail to
    # be redfined as a property.

    def __getattr__(self, name):
        try:
            return self._data[name]
        except KeyError as e:
            raise AttributeError(unicode(e))

    def __setattr__(self, name, value):
        if name[0] == '_':
            object.__setattr__(self, name, value)
        else:
            self._data[name] = value

    def __unicode__(self):
        # TRANS: This is a very brief desription of the unique, pertinent
        # information about a killmail.
        return gettext(u"%(kill_id)d: %(pilot)s lost a ship. Verified: "
                       u"%(verified)s",
                kill_id=self.kill_id,
                pilot=self.pilot,
                ship=self.ship,
                verified=self.verified)

    def __iter__(self):
        """Iterate over the attributes of this killmail.

        Yields tuples in the form ``('<name>', <value>)``. This is used by
        :py:meth:`Request.__init__ <evesrp.models.Request.__init__>` to
        initialize its data quickly. ``<name>`` in the returned tuples is the
        name of the attribute on the :py:class:`~evesrp.models.Request`.
        """
        yield ('id', self.kill_id)
        yield ('ship_type', self.ship)
        yield ('corporation', self.corp)
        yield ('alliance', self.alliance)
        yield ('killmail_url', self.url)
        yield ('base_payout', self.value)
        yield ('kill_timestamp', self.timestamp)
        yield ('system', self.system)
        yield ('constellation', self.constellation)
        yield ('region', self.region)
        yield ('pilot_id', self.pilot_id)


    # TRANS: This is a description of the killmail processor. This specific
    # text should not be shown to the user, but should still be localized just
    # in case.
    description = lazy_gettext(u"A generic Killmail. If you see this text, you"
                               u" need to configure your application.")
    """A user-facing description of what kind of killmails this
    :py:class:`Killmail` validates/handles. This text is displayed below
    the text field for a killmail URL to let users know what kinds of links
    are acceptable.
    """


class ShipNameMixin(object):
    """Killmail mixin providing :py:attr:`Killmail.ship` from
    :py:attr:`Killmail.ship_id`.
    """

    @property
    def ship(self):
        """Looks up the ship name using :py:attr:`Killmail.ship_id`.
        """
        return ships.ships[self.ship_id]


class LocationMixin(object):
    """Killmail mixin for providing solar system, constellation and region
    names from :py:attr:`Killmail.system_id`.
    """

    @property
    def system(self):
        """Provides the solar system name using :py:attr:`Killmail.system_id`.
        """
        return systems.system_names[self.system_id]

    @property
    def constellation(self):
        """Provides the constellation name using :py:attr:`Killmail.system_id`.
        """
        constellation_id = systems.systems_constellations[self.system_id]
        return systems.constellation_names[constellation_id]

    @property
    def region(self):
        """Provides the region name using :py:attr:`Killmail.system_id`.
        """
        constellation_id = systems.systems_constellations[self.system_id]
        region_id = systems.constellations_regions[constellation_id]
        return systems.region_names[region_id]


class RequestsSessionMixin(object):
    """Mixin for providing a :py:class:`requests.Session`.

    The shared session allows HTTP user agents to be set properly, and for
    possible connection pooling.

    .. py:attribute:: requests_session

        A :py:class:`~requests.Session` for making HTTP requests.
    """
    def __init__(self, requests_session=None, **kwargs):
        """Set up a :py:class:`~requests.Session` for making HTTP requests.

        If an existing session is not provided, one will be created.

        :param requests_session: an existing session to use.
        :type requests: :py:class:`~requests.Session`
        """
        if requests_session is None:
            try:
                self.requests_session = current_app.requests_session
            except (AttributeError, RuntimeError):
                self.requests_session = requests.Session()
        else:
            self.requests_session = requests_session
        super(RequestsSessionMixin, self).__init__(**kwargs)


class LegacyZKillmail(Killmail, RequestsSessionMixin, ShipNameMixin,
                      LocationMixin):
    """A killmail sourced from a zKillboard based killboard."""

    zkb_regex = re.compile(r'/(detail|kill)/(?P<kill_id>\d+)/?')

    def __init__(self, url, **kwargs):
        """Create a killmail from the given zKillboard URL.

        :param str url: The URL of the killmail.
        :raises ValueError: if ``url`` isn't a valid zKillboard killmail.
        :raises LookupError: if the zKillboard API response is in an unexpected
            format.
        """
        super(LegacyZKillmail, self).__init__(**kwargs)
        self.url = url
        match = self.zkb_regex.search(url)
        if match:
            self.kill_id = int(match.group('kill_id'))
        else:
            raise ValueError(gettext(u"'%(url)s' is not a valid %(source)s "
                                     u"killmail.",
                                     source=u'zKillboard',
                                     url=self.url))
        parsed = urlparse(self.url, scheme='https')
        if parsed.netloc == '':
            # Just in case someone is silly and gives an address without a
            # scheme. Also fix self.url to have a scheme.
            parsed = urlparse('//' + url, scheme='https')
            self.url = parsed.geturl()
        self.domain = parsed.netloc
        # Check API
        api_url = [a for a in parsed]
        api_url[2] = '/api/no-attackers/no-items/killID/{}/'.format(
                self.kill_id)
        resp = self.requests_session.get(urlunparse(api_url))
        # TRANS: Error message shown when there's a problem accessing the
        # zKillboard API.
        retrieval_error = LookupError(gettext(u"Error retrieving killmail "
                                              u"data: %(code)d",
                                              code=resp.status_code))
        if resp.status_code != 200:
            raise retrieval_error
        try:
            json = resp.json(parse_float=Decimal)
        except ValueError as e:
            raise retrieval_error
        try:
            json = json[0]
        except IndexError as e:
            # TRANS: This is an error message when the killmail is somehow
            # failing to be recognized. The %(url)s is replaced with the URL of
            # the offending killmail.
            raise LookupError(gettext(u"Invalid killmail: %(url)s", url=url))
        # JSON is defined to be UTF-8 in the standard
        victim = json[u'victim']
        self.pilot_id = int(victim[u'characterID'])
        self.pilot = victim[u'characterName']
        self.corp_id = int(victim[u'corporationID'])
        self.corp = victim[u'corporationName']
        if victim[u'allianceID'] != '0':
            self.alliance_id = int(victim[u'allianceID'])
            self.alliance = victim[u'allianceName']
        self.ship_id = int(victim[u'shipTypeID'])
        self.system_id = int(json[u'solarSystemID'])
        # For consistency, store self.value in millions. Decimal is being used
        # for precision at large values.
        # Old versions of zKB don't give the ISK value
        self.value = json[u'zkb'].get(u'totalValue', Decimal(0))
        # Old versions of zKB return numerical values as strings.
        if not isinstance(self.value, Decimal):
            self.value = Decimal(self.value)
        # Parse the timestamp
        time_struct = time.strptime(json[u'killTime'], '%Y-%m-%d %H:%M:%S')
        self.timestamp = dt.datetime(*(time_struct[0:6]),
                tzinfo=utc)

    @property
    def verified(self):
        # zKillboard assigns unverified IDs negative numbers
        return self.kill_id > 0

    def __unicode__(self):
        # TRANS: A quick summary of a killmail's pertinent information.
        return gettext(u"%(kill_id)d: %(pilot)s lost a ship. Verified: "
                       u"%(verified)s. ZKillboard URL: %(url).",
                kill_id=self.kill_id,
                pilot=self.pilot,
                ship=self.ship,
                verified=self.verified,
                url=self.url)


    # TRANS: Decscribing the acceptable killmails for the ZKillboard killmail
    # processor.
    description = lazy_gettext(u'A link to a lossmail from <a '
                                      u'href="https://zkillboard.com/">'
                                      u'ZKillboard</a>.')


class ESIMail(Killmail, RequestsSessionMixin, LocationMixin):
    """A killmail with data sourced from a ESI killmail link."""

    url_regex = re.compile(r'/v1/killmails/(?P<kill_id>\d+)/[0-9a-f]+/')

    killmail_source = u'ESI'

    @classmethod
    def _raise_esi_lookup(cls, esi_error):
        # TRANS: The %s here will be replaced with the (non-localized
        # probably) error from CCP. %(source) will be replaced with the source
        # of the killmail, usually either ESI or ZKillboard.
        raise LookupError(gettext(u"Error retrieving %(source)s killmail: "
                                  u"%(error)s",
                                  source=cls.killmail_source,
                                  error=esi_error))

    def __init__(self, url, **kwargs):
        """Create a killmail from a ESI killmail link.

        :param str url: the ESI killmail URL.
        :raises ValueError: if ``url`` is not a ESI URL.
        :raises LookupError: if the ESI API response is in an unexpected
            format.
        """
        super(ESIMail, self).__init__(**kwargs)
        self.url = url
        match = self.url_regex.search(self.url)
        if match:
            self.kill_id = match.group('kill_id')
        else:
            # TRANS: The %(url)s in this case will be replaced with the
            # offending URL. %(source) will be replaced with the source of the
            # killmail (usually either ESI or zKillboard).
            raise ValueError(gettext(u"'%(url)s' is not a valid %(source)s "
                                     u"killmail.",
                                     source=self.killmail_source,
                                     url=self.url))
        parsed = urlparse(self.url, scheme='https')
        if parsed.netloc == '':
            parsed = urlparse('//' + url, scheme='https')
            self.url = parsed.geturl()
        # Check if it's a valid ESI URL
        resp = self.requests_session.get(self.api_url)
        # JSON responses are defined to be UTF-8 encoded
        # TODO handle all the documented error codes form CCP
        if resp.status_code != 200:
            try:
                error_message = resp.json()[u'error']
            except ValueError:
                error_message = str(resp.status_code)
            self._raise_esi_lookup(error_message)
        try:
            json = resp.json()
        except ValueError as e:
            # TRANS: The %(code)d in this message will be replaced with the
            # HTTP status code recieved from CCP.
            raise LookupError(gettext(u"Error retrieving killmail data: "
                                      u"%(code)d", code=resp.status_code))
        self.ingest_killmail(json)

    @property
    def api_url(self):
        # ESI uses the same URL for the API and to identify this killmail
        return self.url

    def ingest_killmail(self, json_response):
        victim = json_response[u'victim']
        self.pilot_id = victim[u'character_id']
        self.corp_id = victim[u'corporation_id']
        self.alliance_id = victim.get(u'alliance_id')
        self.ship_id = victim[u'ship_type_id']
        self.system_id = json_response[u'solar_system_id']
        self.timestamp = iso8601.parse_date(json_response[u'killmail_time'])
        # Look up the names of the *_id attributes above. Doing this in here
        # instead of making a mixin because /universe/names lets us cut down on
        # the number of requests we have to make.
        ids = [self.ship_id, self.system_id, self.pilot_id, self.corp_id]
        if self.alliance_id not in (0, None):
            # Handle corporations not in alliances
            ids.append(self.alliance_id)
        names_resp = self.requests_session.post(
            'https://esi.tech.ccp.is/v2/universe/names/', json=ids)
        if names_resp.status_code != 200:
            self._raise_esi_lookup(names_resp.json()[u'error'])
        for name_info in names_resp.json():
            category_attributes = {
                'solar_system': 'system',
                'inventory_type': 'ship',
                'character': 'pilot',
                'alliance': 'alliance',
                'corporation': 'corp',
            }
            attr_name = category_attributes[name_info[u'category']]
            name = name_info[u'name']
            setattr(self, attr_name, name)
        # ESI Killmails are always verified
        self.verified = True

    @classproperty
    def description(cls):
        # TRANS: Description of the allowable links for the ESI killmail
        # processor. %(source) is where the killmal is from, usually either ESI
        # or zKillboard.
        return gettext(u'A %(source)s external killmail link.',
                       source=cls.killmail_source)


class ZKillmail(ESIMail):

    url_regex = re.compile(r'/kill/(?P<kill_id>\d+)/?')

    killmail_source = u'zKillboard'

    @property
    def api_url(self):
        # We use a minimized API URL to query against zKB
        return ('https://zkillboard.com/'
                'api/no-attackers/no-items/killID/{}/').format(self.kill_id)

    def ingest_killmail(self, json_response):
        # zKillboard has the ability to return multiple killmails in one API
        # response, so the response is an array of killmail objects, while ESI
        # only returns one killmail per response.
        try:
            killmail_json = json_response[0]
        except IndexError as idx_exc:
            val_exc = ValueError(gettext(u"'%(url)s' is not a valid %(source)s"
                                         u" killmail.",
                                         source=self.killmail_source,
                                         url=self.url))
            six.raise_from(val_exc, idx_exc)
        super(ZKillmail, self).ingest_killmail(killmail_json)
        # Continue on and add some zKB-specific info
        # TODO: customize JSON parser to pull out these values as decimals, not
        # floats
        self.value = Decimal(killmail_json[u'zkb'][u'totalValue'])


# Backwards compatibility
CRESTMail = ESIMail



from __future__ import absolute_import

import xml.etree.ElementTree as ET
from flask import request, current_app, abort, flash
from sqlalchemy.orm.exc import NoResultFound
from flask_wtf import Form
from flask_login import current_user
from oauthlib.oauth2 import OAuth2Error

from .oauth import OAuthMethod, OAuthUser
from .. import db
from .models import Group, Pilot
from ..util.fields import ImageField
from ..util.crest import check_crest_response
from ..versioned_static import static_file


class EveSSO(OAuthMethod):

    def __init__(self, singularity=False, **kwargs):
        kwargs.setdefault('base_url', u'')
        if singularity:
            self.domain = 'https://sisilogin.testeveonline.com'
            self.xml_root = 'https://api.testeveonline.com/'
        else:
            self.domain = 'https://login.eveonline.com'
            self.xml_root = 'https://api.eveonline.com/'
        kwargs.setdefault('access_token_url', self.domain + '/oauth/token')
        kwargs.setdefault('authorize_url', self.domain + '/oauth/authorize')
        kwargs.setdefault('refresh_token_url', self.domain + '/oauth/token')
        kwargs.setdefault('scope', ['publicData'])
        kwargs.setdefault('method', 'POST')
        kwargs.setdefault('app_key', 'EVE_SSO')
        kwargs.setdefault('name', u'EVE SSO')
        super(EveSSO, self).__init__(**kwargs)

    def _get_user_data(self):
        if not hasattr(request, '_user_data'):
            try:
                resp = self.session.get(self.domain + '/oauth/verify').json()
                current_app.logger.debug(u"SSO lookup results: {}".format(resp))
            except OAuth2Error as e:
                current_app.logger.error(u"Error verifying user data for user "
                                         u"'{}': {}".format(current_user, e))
                # The session can be bugged in some situations. Kill it to be
                # sure.
                del self.session
                raise
            try:
                char_data = {
                    'name': resp[u'CharacterName'],
                    'id': resp[u'CharacterID'],
                    'owner_hash': resp[u'CharacterOwnerHash'],
                }
                request._user_data = char_data
            except (TypeError, KeyError):
                abort(500, u"Error in receiving EVE SSO response: {}".format(
                        resp))
        return request._user_data

    def form(self):
        class EveSSOForm(Form):
            submit = ImageField(src=static_file('evesso.png'),
                    alt=u"Log in with EVE Online")

        return EveSSOForm

    def get_user(self):
        character = self._get_user_data()
        try:
            user = EveSSOUser.query.filter_by(
                    owner_hash=character['owner_hash'],
                    authmethod=self.name).one()
        except NoResultFound:
            user = EveSSOUser(
                    character['name'],
                    character['owner_hash'],
                    self.name)
            db.session.add(user)
            db.session.commit()
        return user

    def get_pilots(self):
        # The EVE SSO API only authenticates one character at a time, so we're
        # going to have a 1-to-1 mapping of Users to Pilots
        character = self._get_user_data()
        pilot = Pilot.query.get(int(character['id']))
        if pilot is None:
            pilot = Pilot(None, character['name'], character['id'])
            db.session.add(pilot)
            db.session.commit()
        return [pilot]

    def get_groups(self):
        """Set the user's groups for their pilot.

        At this time, Eve SSO only gives us character access, so they're just
        set to the pilot's corporation, and if they have on their alliance as
        well. In the future, this method may also add groups for mailing lists.
        """
        character = self._get_user_data()
        info_url = self.xml_root + 'eve/CharacterInfo.xml.aspx'
        info_response = current_app.requests_session.get(info_url,
                params={'characterID': character['id']})
        api_tree = ET.fromstring(info_response.text).find('result')
        corp_name = api_tree.find('corporation')
        corp_id = api_tree.find('corporationID')
        corporation = {
            'name': corp_name.text,
            'id': int(corp_id.text),
        }
        try:
            corp_group = EveSSOGroup.query.filter_by(alliance=False,
                    ccp_id=int(corp_id.text),
                    authmethod=self.name).one()
        except NoResultFound:
            corp_group = EveSSOGroup(corp_name.text, int(corp_id.text), False,
                    self.name)
            db.session.add(corp_group)
        groups = [corp_group]

        alliance_name = api_tree.find('alliance')
        alliance_id = api_tree.find('allianceID')
        # If there's an alliance, set it up
        if alliance_name is not None and alliance_id is not None:
            try:
                alliance_group = EveSSOGroup.query.filter_by(alliance=True,
                        ccp_id=int(alliance_id.text),
                        authmethod=self.name).one()
            except NoResultFound:
                alliance_group = EveSSOGroup(alliance_name.text,
                        int(alliance_id.text), True, self.name)
                db.session.add(alliance_group)
            groups.append(alliance_group)
        db.session.commit()
        return groups


class EveSSOUser(OAuthUser):

    id = db.Column(db.Integer, db.ForeignKey(OAuthUser.id), primary_key=True)

    owner_hash = db.Column(db.String(50), nullable=False, unique=True,
            index=True)

    def __init__(self, username, owner_hash, authmethod, groups=None, **kwargs):
        self.owner_hash = owner_hash
        super(EveSSOUser, self).__init__(username, authmethod, **kwargs)


class EveSSOGroup(Group):

    id = db.Column(db.Integer, db.ForeignKey(Group.id), primary_key=True)

    ccp_id = db.Column(db.Integer, nullable=False, unique=True, index=True)

    alliance = db.Column(db.Boolean(name='alliance'), nullable=False, default=False,
            index=True)

    __table_args__ = (
            db.UniqueConstraint(ccp_id, alliance, name='alliance_ccp_id'),
    )

    def __init__(self, name, ccp_id, alliance, authmethod, **kwargs):
        self.ccp_id = ccp_id
        self.alliance = alliance
        super(EveSSOGroup, self).__init__(name, authmethod, **kwargs)



from __future__ import absolute_import
from base64 import urlsafe_b64decode
import binascii
from flask import render_template, url_for, abort, session, redirect, request,\
        current_app, g, Blueprint, flash
import flask_login
from flask_babel import gettext, lazy_gettext
from flask_wtf import Form
from six.moves import map
from wtforms.fields import HiddenField
from wtforms.validators import AnyOf
from sqlalchemy.orm.exc import NoResultFound
from .. import csrf, db
from ..auth import AnonymousUser
from ..auth.models import User, APIKey
from ..util import ensure_unicode, jsonify, xmlify


blueprint = Blueprint('login', __name__)


login_manager = flask_login.LoginManager()
login_manager.anonymous_user = AnonymousUser


@login_manager.user_loader
def login_loader(userid):
    """Pull a user object from the database.

    This is used for loading users from existing sessions.
    """
    return User.query.get(int(userid))


@login_manager.request_loader
def apikey_loader(request):
    api_key = ensure_unicode(request.values.get('apikey'))
    if api_key and request.method == 'GET':
        api_key = api_key.replace(u',', u'=')
        try:
            api_key = urlsafe_b64decode(api_key.encode('utf-8'))
        except binascii.Error:
            # If the api key is malformed, binascii throws an exception.
            # Rejected.
            return None
        try:
            key = APIKey.query.filter_by(key=api_key).one()
        except NoResultFound:
            pass
        else:
            return key.user

    # returning None signifies failure for this method
    return None

@login_manager.needs_refresh_handler
def refresh_user():
    auth_methods = {am.name: am for am in current_app.auth_methods}
    user_auth_method = auth_methods[flask_login.current_user.authmethod]
    if user_auth_method.refresh(flask_login.current_user):
        current_app.logger.debug("Marking '{}' as fresh".format(
            flask_login.current_user))
        flask_login.confirm_login()
        # Call the original endpoint
        view = current_app.view_functions[request.endpoint]
        return view(**request.view_args)
    else:
        flash(login_manager.needs_refresh_message,
                category=login_manager.needs_refresh_message_category)
        original_url = url_for(request.endpoint, **request.view_args)
        return redirect(url_for('login.login', next=original_url,
                _anchor=user_auth_method.safe_name))


class APIKeyForm(Form):
    action = HiddenField(validators=[AnyOf(['add', 'delete'])])
    key_id = HiddenField()


@blueprint.route('/apikeys/', methods=['GET', 'POST'])
@flask_login.login_required
def api_keys():
    form = APIKeyForm()
    if form.validate_on_submit():
        if form.action.data == 'add':
            key = APIKey(flask_login.current_user)
        else:
            key = APIKey.query.get(int(form.key_id.data))
            if key is not None:
                db.session.delete(key)
        db.session.commit()
    if request.is_json or request.is_xhr:
        return jsonify(api_keys=flask_login.current_user.api_keys)
    if request.is_xml:
        return xmlify('apikeys.xml')
    return render_template('apikeys.html', form=form)


@blueprint.route('/login/', methods=['GET', 'POST'])
def login():
    """Presents the login form and processes responses from that form.

    When a POST request is recieved, this function passes control to the
    appropriate :py:meth:`login <evesrp.auth.AuthMethod.login>` method.
    """
    # forms is a list of tuples. The tuples are
    # (AuthMethod instance, AuthForm instance)
    forms = []
    for auth_method in current_app.auth_methods:
        prefix = auth_method.safe_name
        form = auth_method.form()
        forms.append((auth_method, form(prefix=prefix)))
    if request.method == 'POST':
        # Find the form that was submitted. The unique prefix for each form
        # means only one form.submit is going to be valid.
        for auth_tuple in forms:
            if auth_tuple[1].submit.data:
                auth_method, form = auth_tuple
                break
        else:
            abort(400)
        if form.validate():
            return auth_method.login(form)
    return render_template('login.html', forms=forms,
            title=lazy_gettext(u'Log In'))


def localize_login_messages(message):
    return gettext(message)


login_manager.login_view = 'login.login'
login_manager.localize_callback = localize_login_messages


# 302 redirects let the request method change to GET if it started as POST.
# By defining routes for both of these paths, the implicit redirect for the
# first route is skipped.
@blueprint.route('/login/<string:auth_method>', methods=['GET', 'POST'])
@blueprint.route('/login/<string:auth_method>/', methods=['GET', 'POST'])
def auth_method_login(auth_method):
    """Trampoline for :py:class:`~evesrp.auth.AuthMethod`\-specific views.

    See :py:meth:`Authmethod.view <evesrp.auth.AuthMethod.view>` for more
    details.
    """
    method_map = dict(map((lambda m: (m.safe_name, m)),
        current_app.auth_methods))
    return method_map[auth_method].view()


@blueprint.route('/logout/')
@flask_login.login_required
def logout():
    """Logs the current user out.

    Redirects to :py:func:`.index`.
    """
    flask_login.logout_user()
    return redirect(url_for('index'))



from flask import render_template, make_response


def xmlify(template, content_type='application/xml', **context):
    """Render a template and modify the ``Content-Type`` header to say it's XML

    The ``template`` and ``context`` arguemnts are identical to those of
    :py:func:`flask.render_template`.

    :param str content_type: The MIME type that the ``Content-Type`` header is
        to be set to. Defaults to ``'application/xml'``.
    """
    content = render_template(template, **context)
    response = make_response(content)
    response.headers['Content-Type'] = content_type
    return response



from __future__ import absolute_import
from collections import OrderedDict, defaultdict
import re

import babel
from flask import render_template, abort, url_for, flash, Markup, request,\
    redirect, current_app, Blueprint, Markup, json, make_response
from flask.views import View
from flask_babel import gettext, lazy_gettext, get_locale
from flask_login import login_required, fresh_login_required, \
    current_user, login_fresh
from flask_sqlalchemy import Pagination
from flask_wtf import Form
import iso8601
import six
from six.moves import map
from wtforms.fields import SelectField, SubmitField, TextAreaField, HiddenField
from wtforms.fields.html5 import URLField, DecimalField
from wtforms.validators import InputRequired, AnyOf, URL, ValidationError,\
    StopValidation

from .login import login_manager
from .. import db
from ..models import Request, Modifier, Action, ActionType, ActionError,\
        ModifierError, AbsoluteModifier, RelativeModifier
from ..util import xmlify, jsonify, classproperty, PrettyDecimal, varies,\
        ensure_unicode, parse_datetime
from ..auth import PermissionType
from ..auth.models import Division, Pilot, Permission, User, Group, Note,\
    APIKey, users_groups


if six.PY3:
    unicode = str


blueprint = Blueprint('requests', __name__)


class RequestListing(View):
    """Abstract class for lists of :py:class:`~evesrp.models.Request`\s.

    Subclasses will be able to respond to both normal HTML requests as well as
    to API requests with JSON.
    """

    #: The template to use for listing requests
    template = 'requests_list.html'

    #: Decorators to apply to the view functions
    decorators = [login_required, varies('Accept', 'X-Requested-With')]

    # TRANS: The default title for a page with atable listing SRP requests.
    title = lazy_gettext(u'Requests')

    @staticmethod
    def parse_filter(filter_string):
        filters = {}
        # Fail early for empty filters
        if filter_string is None or filter_string == '':
            return filters
        split_filters = filter_string.split('/')
        # Trim empty beginnings and/or ends
        if split_filters[0] == '':
            split_filters = split_filters[1:]
        if split_filters[-1] == '':
            split_filters = split_filters[:-1]
        # Check for unpaired filters
        if len(split_filters) % 2 != 0:
            # FIXME: uneven length of filters
            return filters
        for i in range(0, len(split_filters), 2):
            attr = split_filters[i].lower()
            values = split_filters[i + 1]
            # Use sets for deduplicating
            filters.setdefault(attr, set())
            # Details are special filter types, and may contain commas. They
            # are allowed to be specified multiple times.
            if attr == 'details':
                filters[attr].add(values)
            elif attr == 'page':
                def flash_page_num_warning(page_num):
                    # TRANS: Warning message shown if something other than a
                    # TRANS: single number was given as a page number (like a
                    # TRANS: letter, word, or multiple numbers).
                    flash(gettext(
                            u"Invalid value for page number: %(num)d.",
                                num=page_num),
                            u'warning')
                    current_app.logger.warn(
                            "Invalid value of page number: {}".format(
                                page_num))
                # Check for things that are not page numbers being given as a
                # page number.
                try:
                    filters['page'] = int(values)
                except TypeError:
                    flash_page_num_warning(values)
                # Flash a warning about using a 0 or negative page number, and
                # clamp it to 1
                if filters['page'] < 1:
                    flash_page_num_warning(filters['page'])
                    filters['page'] = 1
            elif attr == 'sort':
                filters['sort'] = values.lower()
            elif attr == 'status':
                actions = set()
                if ',' in values:
                    values = values.split(',')
                else:
                    values = [values]
                for action in values:
                    actions.add(ActionType.from_string(action))
                filters[attr] = actions
            elif ',' in values:
                values = values.split(',')
                filters[attr].update(values)
            else:
                filters[attr].add(values)
        return filters

    @staticmethod
    def unparse_filter(filters):
        attrs = sorted(six.iterkeys(filters))
        filter_strings = []
        for attr in attrs:
            if attr == 'details':
                for details in sorted(filters[attr]):
                    filter_strings.append('details/' + details)
            elif attr == 'page':
                filter_strings.append('page/{}'.format(filters['page']))
            elif attr == 'sort':
                filter_strings.append('sort/{}'.format(filters['sort']))
            elif attr == 'status':
                values = [a.name for a in filters[attr]]
                values = sorted(values)
                filter_strings.append(attr + '/' + ','.join(values))
            else:
                values = sorted(filters[attr])
                filter_strings.append(attr + '/' + ','.join(values))
        return '/'.join(filter_strings)

    def requests(self, filters):
        """Returns a list :py:class:`~.Request`\s belonging to
        the specified :py:class:`~.Division`, or all divisions if
        ``None``.

        :returns: :py:class:`~.models.Request`\s
        :rtype: iterable
        """
        # Start with a basic query for requests
        requests = Request.query.options(*self._load_options)
        requests = requests.order_by(Request.timestamp.desc())
        # Set default filters values
        filters.setdefault('page', 1)
        filters.setdefault('sort', '-submit_timestamp')
        # Apply the filters
        known_attrs = ('page', 'division', 'alliance', 'corporation',
                'pilot', 'system', 'constellation', 'region', 'ship_type',
                'status', 'details', 'payout', 'base_payout', 'kill_timestamp',
                'submit_timestamp')
        for attr, values in six.iteritems(filters):
            # massage pretty attribute names to the not-so-pretty ones
            if attr == 'ship':
                real_attr = 'ship_type'
            elif attr == 'submit_timestamp':
                real_attr = 'timestamp'
            else:
                real_attr = attr
            # massage negative/range filters
            if real_attr not in ('page', 'sort', 'details', 'status',
                                 'timestamp', 'kill_timestamp'):
                new_values = set()
                for value in values:
                    if value.startswith(('-', '<', '>')):
                        new_values.add((value[:1], value[1:]))
                    elif value.startswith(('<=', '>=')):
                        new_values.add((value[:1], value[:2]))
                        new_values.add(('=', value[:2]))
                    else:
                        new_values.add(('=', value))
                grouped = {'=': set(), '<': set(), '>': set(), '-': set()}
                for value in new_values:
                    grouped[value[0]].add(value[1])
            else:
                grouped = {'=': values, '-':[], '<':[], '>':[]}
            # Handle a couple attributes specially
            if real_attr == 'details':
                clauses = [Request.details.match(d) for d in values]
                requests = requests.filter(db.or_(*clauses)) 
            elif real_attr == 'page':
                continue
            elif real_attr == 'sort':
                if values[0] == '-':
                    descending = True
                    sort_attr = values[1:]
                else:
                    descending = False
                    sort_attr = values
                # massage special attribute names
                if sort_attr == 'ship':
                    sort_attr = 'ship_type'
                elif sort_attr == 'submit_timestamp':
                    sort_attr = 'timestamp'
                # Handle special (joined) sorts
                if sort_attr == 'division':
                    if descending:
                        column = db.func.lower(Division.name).desc()
                    else:
                        column = db.func.lower(Division.name).asc()
                    requests = requests.order_by(None)
                    requests = requests.join(Division).order_by(column)
                elif sort_attr == 'pilot':
                    if descending:
                        column = db.func.lower(Pilot.name).desc()
                    else:
                        column = db.func.lower(Pilot.name).asc()
                    requests = requests.order_by(None)
                    requests = requests.join(Pilot).order_by(column)
                else:
                    column = getattr(Request, sort_attr)
                    if descending:
                        column = column.desc()
                    else:
                        column = column.asc()
                    requests = requests.order_by(None)
                    requests = requests.order_by(column)
            elif real_attr in ('timestamp', 'kill_timestamp'):
                column = getattr(Request, real_attr)
                for value in values:
                    try:
                        start, end = parse_datetime(value)
                    except iso8601.ParseError as e:
                        current_app.logger.error(
                                u"Invalid date provided ({}). "
                                u"Exception: {}".format(value, e))
                        # TRANS: Error message shown when a date is written
                        # TRANS: incorrectly.
                        abort(400, gettext(u"Invalid date format. Please read "
                                           u"the documentation for date "
                                           u"filtering."))
                    requests = requests.filter(column.between(start, end))
            elif real_attr in known_attrs:
                column = getattr(Request, real_attr)
                # in_ isn't supported on relationships (yet).
                if hasattr(column, 'mapper'):
                    # This is black magic
                    id_column = column.mapper.attrs.id.class_attribute
                    name_column = column.mapper.attrs.name.class_attribute
                    mapped = db.session.query(id_column)
                    filtered = False
                    if grouped['=']:
                        mapped = mapped.filter(name_column.in_(grouped['=']))
                        filtered = True
                    if grouped['-']:
                        mapped = mapped.filter(~name_column.in_(grouped['-']))
                        filtered = True
                    if grouped['<']:
                        for lt_val in grouped['<']:
                            mapped = mapped.filter(name_column < lt_val)
                        filtered = True
                    if grouped['>']:
                        for gt_val in grouped['>']:
                            mapped = mapped.filter(name_column > gt_val)
                        filtered = True
                    if filtered:
                        mapped = mapped.subquery()
                        requests = requests.join(mapped)
                else:
                    if grouped['=']:
                        requests = requests.filter(column.in_(grouped['=']))
                    if grouped['-']:
                        requests = requests.filter(~column.in_(grouped['-']))
                    for lt_val in grouped['<']:
                        requests = requests.filter(column < lt_val)
                    for gt_val in grouped['>']:
                        requests = requests.filter(column > gt_val)
            else:
                # TRANS: Warning message shown when an unknown atribute is
                # TRANS: specified in the list of filters.
                flash(gettext(
                        u"Unknown filter attribute name: %(attribute)s.",
                        attribute=attr),
                    u'warning')
        return requests

    def dispatch_request(self, filters='', **kwargs):
        """Returns the response to requests.

        Part of the :py:class:`flask.views.View` interface.
        """
        filter_map = self.parse_filter(filters)
        current_app.logger.debug("Filter map: {}".format(filter_map))
        canonical_filter = self.unparse_filter(filter_map)
        if canonical_filter != filters:
            current_app.logger.debug(u"Redirecting to filter '{}' from filter"
                                     u" '{}'.".format(canonical_filter,
                                         filters))
            url_kwargs = {
                'filters': canonical_filter,
            }
            if 'fmt' in request.args:
                url_kwargs['fmt'] = request.args['fmt']
            return redirect(url_for(request.endpoint, **url_kwargs), code=301)
        requests = self.requests(filter_map)
        # Ignore rejected requests when summing the payout.
        # Discard ordering options, they affect the sum somehow.
        payout_requests = requests.\
                filter(Request.status != ActionType.rejected).\
                order_by(False).\
                with_entities(Request.id, Request.payout).\
                subquery(with_labels=True)
        total_payouts = db.session.query(db.func.sum(
                        payout_requests.c.request_payout))\
                .select_from(payout_requests)\
                .scalar()
        if total_payouts is None:
            total_payouts = PrettyDecimal(0)
        # API requests (including RSS) should have a limit of 200 requests,
        # otherwise go with the standard 15. THe standard 15 includes
        # XmlHttpRequest-based requests as those are going to be used to
        # rebuild the contents of an HTML view.
        if (request.is_json or request.is_xml or request.is_rss) and \
                not request.is_xhr:
            per_page = 200
        else:
            per_page = 15
        pager = requests.paginate(filter_map['page'], per_page=per_page,
                error_out=False)
        # Redirect to the last page if the request page is larger than the
        # number of pages.
        if len(pager.items) == 0 and pager.page > 1:
            filter_map['page'] = pager.pages
            return redirect(url_for(request.endpoint,
                    filters=self.unparse_filter(filter_map)))
        # Prep previous and next page links for API responses
        api_links = {}
        link_kwargs = {
            '_external': True,
        }
        if 'fmt' in request.args:
            link_kwargs['fmt'] = request.args['fmt']
        if pager.has_prev:
            filter_map['page'] -= 1
            api_links['prev'] = url_for(request.endpoint,
                    filters=self.unparse_filter(filter_map), **link_kwargs)
            filter_map['page'] = pager.page
        if pager.has_next:
            filter_map['page'] += 1
            api_links['next'] = url_for(request.endpoint,
                    filters=self.unparse_filter(filter_map), **link_kwargs)
            filter_map['page'] = pager.page
        # Handle API/RSS responses
        if request.is_json or request.is_xhr:
            jsonify_kwargs = {
                'requests': pager.items,
                'request_count': requests.count(),
                'total_payouts': total_payouts.currency()
            }
            # For JSON responses, add prev and next links (when appropriate) to
            # aid API consumers in walking a list of requests.
            if request.is_json:
                jsonify_kwargs.update(api_links)
            return jsonify(**jsonify_kwargs)
        if request.is_rss:
            return xmlify('rss.xml', content_type='application/rss+xml',
                    requests=pager.items,
                    title=(kwargs['title'] if 'title' in kwargs else u''),
                    main_link=url_for(request.endpoint, filters=filters,
                        _external=True))
        if request.is_xml:
            xmlify_kwargs = {
                'requests': pager.items,
                'total_payouts': total_payouts,
            }
            # As for JSON API responses, add prev and next links
            xmlify_kwargs.update(api_links)
            return xmlify('requests_list.xml', **xmlify_kwargs)
        if 'title' in kwargs:
            title = kwargs.pop('title')
        else:
            title = self.title
        return render_template(self.template, pager=pager, filters=filter_map,
                total_payouts=total_payouts, title=title, **kwargs)

    @classproperty
    def _load_options(cls):
        """Returns a sequence of
        :py:class:`~sqlalchemy.orm.strategy_options.Load` objects specifying
        which attributes to load (or really any load options necessary).
        """
        return (
                db.Load(Request).load_only('id', 'pilot_id', 'division_id',
                    'system', 'ship_type', 'status', 'timestamp',
                    'base_payout'),
                db.Load(Division).joinedload('name'),
                db.Load(Pilot).joinedload('name'),
        )


def url_for_page(pager, page_num):
    """Utility method used in Jinja templates."""
    filters = request.view_args.get('filters', '')
    filters = RequestListing.parse_filter(filters)
    filters['page'] = page_num
    return url_for(request.endpoint,
            filters=RequestListing.unparse_filter(filters))


class PersonalRequests(RequestListing):
    """Shows a list of all personally submitted requests and divisions the user
    has permissions in.

    It will show all requests the current user has submitted.
    """

    template = 'requests_personal.html'

    decorators = [login_required]

    def requests(self, filters):
        requests = super(PersonalRequests, self).requests(filters)
        requests = requests\
                .join(User)\
                .filter(User.id==current_user.id)
        return requests

    @property
    def title(self):
        current_locale = get_locale()
        if current_locale is None:
            current_locale = babel.Locale('en')
        # Special case possesive form in English
        if current_locale.language.lower() == 'en' and \
                current_user.name[:-1] == u's':
            return u"{}' Requests".format(current_user.name)
        else:
            # TRANS: The title of the page listing all requests an individual
            # TRANS: user has made.
            return gettext(u"%(name)s's Requests", name=current_user.name)


class PermissionRequestListing(RequestListing):
    """Show all requests that the current user has permissions to access.

    This is used for the various permission-specific views.
    """

    decorators = [fresh_login_required]

    def __init__(self, permissions, statuses, title=None):
        """Create a :py:class:`PermissionRequestListing` for the given
        permissions and statuses.

        :param tuple permissions: The permissions to filter by
        :param tuple statuses: A tuple of valid statuses for requests to be in
        """
        if permissions in PermissionType.all:
            permissions = (permissions,)
        # Admin permission has to be explicitly added because it's used in a
        # complicated query in requests()
        self.permissions = (PermissionType.admin,) + tuple(permissions)
        self.statuses = statuses
        if title is None:
            self.title = u', '.join(map(lambda s: s.description, self.statuses))
        else:
            self.title = ensure_unicode(title)

    def dispatch_request(self, filters='', **kwargs):
        if not current_user.has_permission(self.permissions):
            abort(403)
        else:
            return super(PermissionRequestListing, self).dispatch_request(
                    filters,
                    title=gettext(self.title),
                    **kwargs)

    def requests(self, filters):
        current_groups = db.select([users_groups.c.group_id])\
                .where(users_groups.c.user_id == current_user.id).alias()
        divisions = db.select([Permission.division_id.label('division_id')])\
                .where(Permission.permission.in_(self.permissions))\
                .where(db.or_(
                        Permission.entity_id == current_user.id,
                        Permission.entity_id.in_(current_groups)))\
                .alias('permitted_divisions')
        # modify filters
        if 'status' not in filters:
            filters['status'] = self.statuses
        requests = super(PermissionRequestListing, self).requests(filters)\
                .join(divisions, Request.division_id==divisions.c.division_id)
        return requests.distinct()


class PayoutListing(PermissionRequestListing):
    """A special view made for quickly processing payouts for requests."""

    template = 'payout.html'

    def __init__(self):
        # Just a special case of PermissionRequestListing
        super(PayoutListing, self).__init__((PermissionType.pay,),
                (ActionType.approved,), u'Pay Outs')

    def requests(self, filters):
        if 'sort' not in filters:
            filters['sort'] = 'submit_timestamp'
        return super(PayoutListing, self).requests(filters)

    def dispatch_request(self, filters='', **kwargs):
        if hasattr(request, 'json_extended'):
            if isinstance(request.json_extended, bool):
                old_value = request.json_extended
                request.json_extended = defaultdict(lambda: old_value)
        else:
            request.json_extended = {}
        request.json_extended[Request] = True
        if not current_user.has_permission(self.permissions):
            abort(403)
        return super(PayoutListing, self).dispatch_request(
                filters,
                form=ActionForm())


def register_perm_request_listing(app, endpoint, path, permissions, statuses,
        title=None):
    """Utility function for creating :py:class:`PermissionRequestListing`
    views.

    :param app: The application to add the view to
    :type app: :py:class:`flask.Flask`
    :param str endpoint: The name of the view
    :param str path: The URL path for the view
    :param tuple permissions: Passed to
        :py:meth:`PermissionRequestListing.__init__`
    :param iterable statuses: Passed to
        :py:meth:`PermissionRequestListing.__init__`
    """
    if not path.endswith('/'):
        path += '/'
    view = PermissionRequestListing.as_view(endpoint, permissions=permissions,
            statuses=statuses, title=title)
    app.add_url_rule(path, view_func=view)
    app.add_url_rule('{}rss.xml'.format(path), view_func=view)
    app.add_url_rule(path + '<path:filters>', view_func=view)


@blueprint.record
def register_class_views(state):
    """Called when the blueprint is registered, this function defines routes
    for and attaches the class-based views to the app.
    """
    try:
        prefixes = state.app.request_prefixes
    except AttributeError:
        prefixes = []
        state.app.request_prefixes = prefixes
    prefixes.append(state.url_prefix if state.url_prefix is not None else '')
    # Personal list
    personal_view = PersonalRequests.as_view('personal_requests')
    state.add_url_rule('/personal/', view_func=personal_view)
    state.add_url_rule('/personal/rss.xml', view_func=personal_view)
    state.add_url_rule('/personal/<path:filters>', view_func=personal_view)
    # Payout list
    payout_view = PayoutListing.as_view('list_approved_requests')
    payout_url_stub = '/pay/'
    state.add_url_rule(payout_url_stub, view_func=payout_view)
    state.add_url_rule(payout_url_stub + 'rss.xml', view_func=payout_view)
    state.add_url_rule(payout_url_stub + '<path:filters>',
            view_func=payout_view)
    # Other more generalized listings
    register_perm_request_listing(state, 'list_pending_requests',
            '/pending/', (PermissionType.review, PermissionType.audit),
            ActionType.pending, u'Pending Requests')
    register_perm_request_listing(state, 'list_completed_requests',
            '/completed/', PermissionType.elevated, ActionType.finalized,
            u'Completed Requests')
    # Special all listing, mainly intended for API users
    register_perm_request_listing(state, 'list_all_requests',
            '/all/', PermissionType.elevated, ActionType.statuses,
            u'All Requests')


class ValidKillmail(URL):
    """Custom :py:class:'~.Field' validator that checks if any
    :py:class:`~.Killmail` accepts the given URL.
    """
    def __init__(self, mail_class, **kwargs):
        self.mail_class = mail_class
        super(ValidKillmail, self).__init__(**kwargs)

    def __call__(self, form, field):
        super(ValidKillmail, self).__call__(form, field)
        try:
            mail = self.mail_class(field.data)
        except ValueError as e:
            if six.PY2:
                raise ValidationError(unicode(e))
            else:
                # Py3 chains the exceptions
                raise ValidationError
        except LookupError as e:
            if six.PY2:
                raise ValidationError(unicode(e))
            else:
                # Py3 chains the exceptions
                raise ValidationError
        else:
            if mail.verified:
                form.killmail = mail
                raise StopValidation
            else:
                raise ValidationError(
                # TRANS: Error message show when trying to submit a killmail
                # TRANS: that cannot be verified.
                        gettext(u"%(url)s cannot be verified.", url=field.data))


def get_killmail_validators():
    """Get a list of :py:class:`ValidKillmail`\s for each killmail source.

    This method is used to delay accessing `current_app` until we're in a
    request context.
    :returns: a list of :py:class:`ValidKillmail`\s
    :rtype list:
    """
    validators = [ValidKillmail(s) for s in current_app.killmail_sources]
    validators.append(InputRequired())
    return validators


def get_killmail_descriptions():
    km_list = u"".join(
            [Markup(u"<li>{}</li>").format(km.description) for km in \
                    current_app.killmail_sources])
    # TRANS: Header for a list of the kinds of links that are acceptable to
    # TRANS: submit as killmails for SRP.
    description = gettext(u"Acceptable Killmail Links:<ul>%(sources)s</ul>",
            sources=km_list)
    description = Markup(description)
    return description


class RequestForm(Form):
    # TRANS: Label for an input field for the killmail URL
    url = URLField(lazy_gettext(u'Killmail URL'))

    # TRANS: Label for a text entry field for inputting supporting details
    # TRANS: about a loss.
    details = TextAreaField(lazy_gettext(u'Details'),
        validators=[InputRequired()])

    # TRANS: Label for an dropdown menu to select which division to submit for
    # TRANS: SRP to.
    division = SelectField(lazy_gettext(u'Division'), coerce=int)

    # TRANS: Label for the button that submits the form.
    submit = SubmitField(lazy_gettext(u'Submit'))

    def validate_url(form, field):
        failures = set()
        for v in get_killmail_validators():
            try:
                v(form, field)
            except ValidationError as e:
                failures.add(unicode(e))
            else:
                continue
        else:
            # If execution reached here, it means a StopValidation exception
            # wasn't raised (meaning the killmail isn't valid).
            raise ValidationError([e for e in failures])

    def validate_division(form, field):
        division = Division.query.get(field.data)
        if division is None:
            # TRANS: Error message shown when trying to submit a request to a
            # TRANS: non-existant division.
            raise ValidationError(gettext(u"No division with ID '%(div_id)s'.",
                    div_id=field.data))
        if not current_user.has_permission(PermissionType.submit, division):
            # TRANS: Error message shown when trying to a submit a request to a
            # TRANS: division you do not have the submission permission in.
            raise ValidationError(gettext(u"You do not have permission to "
                                          u"submit to division '%(name)s'.",
                    name=division.name))


@blueprint.route('/add/', methods=['GET', 'POST'])
@fresh_login_required
def submit_request():
    """Submit a :py:class:`~.models.Request`\.

    Displays a form for submitting a request and then processes the submitted
    information. Verifies that the user has the appropriate permissions to
    submit a request for the chosen division and that the killmail URL given is
    valid. Also enforces that the user submitting this requests controls the
    character from the killmail and prevents duplicate requests.
    """
    if not current_user.has_permission(PermissionType.submit):
        abort(403)
    form = RequestForm()
    # Do it in here so we can access current_app (needs to be in an app
    # context)
    form.url.description = get_killmail_descriptions()
    form.details.description = current_app.config['SRP_DETAILS_DESCRIPTION']
    # Create a list of divisions this user can submit to
    form.division.choices = current_user.submit_divisions()
    if len(form.division.choices) == 1:
        form.division.data = form.division.choices[0][0]

    if form.validate_on_submit():
        mail = form.killmail
        # Prevent submitting other people's killmails
        pilot = Pilot.query.get(mail.pilot_id)
        if not pilot or pilot not in current_user.pilots:
            # TRANS: Error message shown when trying to submit a lossmail from
            # TRANS: A character not associated with your user account.
            flash(gettext(u"You can only submit killmails of characters you "
                          u"control"),
                    u'warning')
            return render_template('form.html', form=form)
        # Prevent duplicate killmails
        # The name 'request' is already used by Flask.
        # Hooray name collisions!
        srp_request = Request.query.get(mail.kill_id)
        if srp_request is None:
            division = Division.query.get(form.division.data)
            srp_request = Request(current_user, form.details.data, division,
                    mail)
            srp_request.pilot = pilot
            db.session.add(srp_request)
            db.session.commit()
            return redirect(url_for('.get_request_details',
                request_id=srp_request.id))
        else:
            # TRANS: Error message shown when trying to submit a killmail for
            # TRANS: SRP a second time.
            flash(gettext(u"This kill has already been submitted"), u'warning')
            return redirect(url_for('.get_request_details',
                request_id=srp_request.id))
    return render_template('form.html', form=form,
    # TRANS: Title for the page showing the form for submitting SRP requests.
            title=gettext(u'Submit Request'))


class ModifierForm(Form):

    id_ = HiddenField(default='modifier')

    # TRANS: Label for an input field for the amount a modifier is for. Can be
    # TRANS: either a percentage or an ISK value (in millions of ISK).
    value = DecimalField(lazy_gettext(u'Value'))

    type_ = HiddenField(validators=[AnyOf(('rel-bonus', 'rel-deduct',
            'abs-bonus', 'abs-deduct'))])

    # TRANS: Label for a text field for inputting the reason a modifier is
    # TRANS: being applied.
    note = TextAreaField(lazy_gettext(u'Reason'))


class VoidModifierForm(Form):

    id_ = HiddenField(default='void')

    modifier_id = HiddenField()

    void = SubmitField(Markup(u'x'))

    def __init__(self, modifier=None, *args, **kwargs):
        if modifier is not None:
            self.modifier_id = modifier.id
        super(VoidModifierForm, self).__init__(*args, **kwargs)


class PayoutForm(Form):

    id_ = HiddenField(default='payout')

    value = DecimalField(u'M ISK', validators=[InputRequired()])


class ActionForm(Form):

    id_ = HiddenField(default='action')

    # TRANS: Label for a text field for adding a note to an action on a
    # TRANS: request, or for making a comment on a request.
    note = TextAreaField(u'Note')

    type_ = HiddenField(default='comment',
            validators=[AnyOf(list(ActionType.values()))])


class ChangeDetailsForm(Form):

    id_ = HiddenField(default='details')

    details = TextAreaField(lazy_gettext(u'Details'),
            validators=[InputRequired()])


class AddNote(Form):

    id_ = HiddenField(default='note')

    # TRANS: Label for text input field for a note to be added about a user.
    note = TextAreaField(lazy_gettext(u'Add Note'),
            # TRANS: Help text explaining a hidden feature about the notes made
            # TRANS: about users.
            description=lazy_gettext(
                u"If you have something like '#{Kill ID}', it will be "
                u"linkified to the corresponding request (if it exists). For "
                u"example, #1234567 would be linked to the request for the "
                u"kill with ID 1234567."),
            validators=[InputRequired()])


@blueprint.route('/<int:request_id>/', methods=['GET'])
@login_required
@varies('Accept')
def get_request_details(request_id=None, srp_request=None):
    """Handles responding to all of the :py:class:`~.models.Request` detail
    functions.

    The various modifier functions all depend on this function to create the
    actual response content.
    Only one of the arguments is required. The ``srp_request`` argument is a
    conveniece to other functions calling this function that have already
    retrieved the request.

    :param int request_id: the ID of the request.
    :param srp_request: the request.
    :type srp_request: :py:class:`~.models.Request`
    """
    if srp_request is None:
        srp_request = Request.query.get_or_404(request_id)
    # A user should always be able to access their own requests, but others
    # need fresh sessions.
    if current_user != srp_request.submitter and not login_fresh():
        return login_manager.needs_refresh()
    # Different templates are used for different roles
    if current_user.has_permission(PermissionType.review,
            srp_request.division):
        template = 'request_review.html'
    elif current_user.has_permission(PermissionType.pay, srp_request.division):
        template = 'request_pay.html'
    elif current_user == srp_request.submitter or current_user.has_permission(
            PermissionType.audit):
        template = 'request_detail.html'
    else:
        abort(403)
    if request.is_json or request.is_xhr:
        return jsonify(**srp_request._json(True))
    if request.is_xml:
        return xmlify('request.xml', srp_request=srp_request)
    return render_template(template, srp_request=srp_request,
            modifier_form=ModifierForm(formdata=None),
            payout_form=PayoutForm(formdata=None),
            action_form=ActionForm(formdata=None),
            void_form=VoidModifierForm(formdata=None),
            details_form=ChangeDetailsForm(formdata=None, obj=srp_request),
            note_form=AddNote(formdata=None),
            # TRANS: Title for the page showing the details about a single
            # TRANS: SRP request.
            title=gettext(u"Request #%(request_id)s",
                    request_id=srp_request.id))


def _add_modifier(srp_request):
    form = ModifierForm()
    if form.validate():
        if 'bonus' in form.type_.data:
            value = form.value.data
        elif 'deduct' in form.type_.data:
            value = form.value.data * -1
        if 'abs' in form.type_.data:
            ModClass = AbsoluteModifier
            value *= 1000000
        elif 'rel' in form.type_.data:
            ModClass = RelativeModifier
            value /= 100
        try:
            mod = ModClass(srp_request, current_user, form.note.data, value)
            db.session.add(mod)
            db.session.commit()
        except ModifierError as e:
            flash(unicode(e), u'error')
    return get_request_details(srp_request=srp_request)


def _change_payout(srp_request):
    form = PayoutForm()
    if not current_user.has_permission(PermissionType.review, srp_request):
        # TRANS: Error message when someone who does not have the reviewer
        # TRANS: permission tries to change the base payout of a request.
        flash(gettext(u"Only reviewers can change the base payout."), u'error')
    elif form.validate():
        try:
            srp_request.base_payout = form.value.data * 1000000
            db.session.commit()
        except ModifierError as e:
            flash(unicode(e), u'error')
    return get_request_details(srp_request=srp_request)


def _add_action(srp_request):
    form = ActionForm()
    if form.validate():
        type_ = ActionType.from_string(form.type_.data)
        try:
            Action(srp_request, current_user, form.note.data, type_)
            db.session.commit()
        except ActionError as e:
            flash(unicode(e), u'error')
    return get_request_details(srp_request=srp_request)


def _void_modifier(srp_request):
    form = VoidModifierForm()
    if form.validate():
        modifier_id = int(form.modifier_id.data)
        modifier = Modifier.query.get(modifier_id)
        if modifier is None:
            # TRANS: Error message when a user tries to void (cancel) a
            # TRANS: modifier that does not exist.
            flash(gettext(u"Invalid modifier ID %(modifier_id)d.",
                    modifier_id=modifier_id),
                u'error')
        else:
            try:
                modifier.void(current_user)
                db.session.commit()
            except ModifierError as e:
                flash(unicode(e), u'error')
    return get_request_details(srp_request=srp_request)


def _change_details(srp_request):
    form = ChangeDetailsForm()
    if current_user != srp_request.submitter:
        # TRANS: Error message shown when someone other than the request
        # TRANS: submitter tries to change the details of the request.
        flash(gettext(u"Only the submitter can change the request details."),
                u'error')
    elif srp_request.finalized:
        # TRANS: Error message shown when the submitter ties to change the
        # TRANS: details when the request is no long pending (not finished).
        flash(gettext(u"Details can only be changed when the request is still "
                      u"pending.")
                , u'error')
    elif form.validate():
        # TRANS: The old request details have are saved in a comment on the
        # TRANS: request. This is the text that is put at the beginning of the
        # TRANS: comment
        archive_note = gettext(u"Old Details: %(details)s",
                details=srp_request.details)
        if srp_request.status == ActionType.evaluating:
            action_type = ActionType.comment
        else:
            action_type = ActionType.evaluating
        archive_action = Action(srp_request, current_user, archive_note,
                action_type)
        srp_request.details = form.details.data
        db.session.commit()
    return get_request_details(srp_request=srp_request)


killmail_re = re.compile(r'#(\d+)')


def _add_note(srp_request):
    form = AddNote()
    if not current_user.has_permission(PermissionType.elevated):
        # TRANS: Error message shown when someone without the proper
        # TRANS: permissions tries to add a note to a user.
        flash(gettext(u"You do not have permission to add a note to a user."),
                u'error')
    elif form.validate():
        # Linkify killmail IDs
        note_content = Markup.escape(form.note.data)
        for match in killmail_re.findall(note_content):
            kill_id = int(match)
            check_request = db.session.query(Request.id).filter_by(id=kill_id)
            if db.session.query(check_request.exists()):
                link = u'<a href="{url}">#{kill_id}</a>'.format(
                        url=url_for('.get_request_details',
                                request_id=kill_id),
                        kill_id=kill_id)
                link = Markup(link)
                note_content = note_content.replace(u'#' + match, link)
        # Create the note
        note = Note(srp_request.submitter, current_user, note_content)
        db.session.commit()
    return get_request_details(srp_request=srp_request)


@blueprint.route('/<int:request_id>/', methods=['POST'])
@login_required
def modify_request(request_id):
    """Handles POST requests that modify :py:class:`~.models.Request`\s.

    Because of the numerous possible forms, this function bounces execution to
    a more specific function based on the form's "id\_" field.

    :param int request_id: the ID of the request.
    """
    srp_request = Request.query.get_or_404(request_id)
    # Force fresh permissions
    if not login_fresh():
        return login_manager.needs_refresh()
    if request.form['id_'] == 'modifier':
        return _add_modifier(srp_request)
    elif request.form['id_'] == 'payout':
        return _change_payout(srp_request)
    elif request.form['id_'] == 'action':
        return _add_action(srp_request)
    elif request.form['id_'] == 'void':
        return _void_modifier(srp_request)
    elif request.form['id_'] == 'details':
        return _change_details(srp_request)
    elif request.form['id_'] == 'note':
        return _add_note(srp_request)
    else:
        return abort(400)


class DivisionChange(Form):

    division = SelectField(lazy_gettext(u'Division'), coerce=int)

    submit = SubmitField(lazy_gettext(u'Submit'))


@blueprint.route('/<int:request_id>/division/', methods=['GET', 'POST'])
@fresh_login_required
def request_change_division(request_id):
    srp_request = Request.query.get_or_404(request_id)
    if not current_user.has_permission(PermissionType.review, srp_request) and\
            current_user != srp_request.submitter:
        current_app.logger.warn(u"User '{}' does not have permission to change"
                                u" request #{}'s division".format(
                                    current_user, srp_request.id))
        abort(403)
    if srp_request.finalized:
        msg = (u"Cannot change request #{}'s division as it is in a finalized"
               u" state").format(srp_request.id)
        current_app.logger.info(msg)
        # TRANS: Error message shown when a user tries to move a request but is
        # TRANS: unable to becuase the request has been paid or rejected.
        flash(gettext(u"Cannot change the division as this request is in a "
                      u"finalized state"),
                u'error')
        return redirect(url_for('.get_request_details', request_id=request_id))
    division_choices = srp_request.submitter.submit_divisions()
    try:
        division_choices.remove(
                (srp_request.division.id, srp_request.division.name))
    except ValueError:
        pass
    if len(division_choices) == 0:
        current_app.logger.debug(u"No other divisions to move request #{} to."\
                .format(srp_request.id))
        # TRANS: Message shown when a user tries to change a request's division
        # TRANS: but they do not have access to any other divisions.
        flash(gettext(u"No other divisions to move to."), u'info')
        return redirect(url_for('.get_request_details', request_id=request_id))
    form = DivisionChange()
    form.division.choices = division_choices
    # Default to the first value if there's only once choice.
    if len(division_choices) == 1:
        form.division.data = form.division.choices[0][0]
    if form.validate_on_submit():
        new_division = Division.query.get(form.division.data)
        # TRANS: When a request is moved from one division to another, a
        # TRANS: comment is added noting the old division's name and the new
        # TRANS: division's name. This is the text of that note.
        archive_note = gettext(u"Moving from division '%(old_division)s' to "
                               u"division '%(new_division)s'.",
                old_division=srp_request.division.name,
                new_division=new_division.name)
        if srp_request.status == ActionType.evaluating:
            type_ = ActionType.comment
        else:
            type_ = ActionType.evaluating
        archive_action = Action(srp_request, current_user, archive_note, type_)
        srp_request.division = new_division
        db.session.commit()
        # TRANS: Confirmation message shown when wa request has been
        # TRANS: successfully moved to a new division.
        flash(gettext(u"Request #%(request_id)d moved to %(division)s "
                      u"division",
                    request_id=srp_request.id, division=new_division.name),
                u'success')
        if current_user.has_permission(PermissionType.elevated, new_division) \
                or current_user == srp_request.submitter:
            return redirect(url_for('.get_request_details',
                    request_id=request_id))
        else:
            return redirect(url_for('.list_pending_requests'))
    else:
        current_app.logger.warn(u"Form validation failed for division change:"
                                u" {}.".format(form.errors))
    form.division.data = srp_request.division.id
    return render_template('form.html', form=form,
            # TRANS: Title for the page showing the form for changing a
            # TRANS: request's division.
            title=lazy_gettext(u"Change #%(request_id)d's Division",
                request_id=srp_request.id))



from flask import request
from six.moves.urllib.parse import urlparse, urljoin


def is_safe_redirect(redirect_url):
    # Fail everything starting with more then one slash
    # http://homakov.blogspot.com/2014/01/evolution-of-open-redirect-vulnerability.html
    if redirect_url.startswith('//'):
        return False
    # Validate given URL to make sure it's still on this server
    current_server = urlparse(request.host_url)
    redirect = urlparse(urljoin(request.host_url, redirect_url))
    return redirect.scheme in ('http', 'https') and \
            redirect.netloc == current_server.netloc



import re
import hashlib
from werkzeug.exceptions import NotFound
from flask import Flask, send_file, safe_join, url_for, g, current_app, abort


def get_file_hash(filename):
    if not hasattr(g, 'static_hashes'):
        g.static_hashes = {}
    # Check the cache if not in debug mode
    if not current_app.debug:
        try:
            return g.static_hashes[filename]
        except KeyError:
            pass
    hasher = hashlib.md5()
    with open(safe_join(current_app.static_folder, filename), 'rb') as f:
        hasher.update(f.read())
    filehash = hasher.hexdigest()[:8]
    g.static_hashes[filename] = filehash
    return filehash


def static_file(filename, **kwargs):
    if current_app.config['SRP_STATIC_FILE_HASH']:
        filehash = get_file_hash(filename)
        beginning, extension = filename.rsplit('.', 1)
        filename = '{}.{}.{}'.format(beginning, filehash, extension)
    return url_for('static', filename=filename, **kwargs)


class VersionedStaticFlask(Flask):

    def send_static_file(self, filename):
        # Short-circuit if not adding file hashes
        if not self.config['SRP_STATIC_FILE_HASH']:
            return super(VersionedStaticFlask, self).send_static_file(filename)
        try:
            return super(VersionedStaticFlask, self).send_static_file(filename)
        except NotFound as e:
            current_app.logger.debug(u"Checking for version-hashed file: {}".
                    format(filename))
            # Map file names are derived from the source file's name, so ignore
            # the '.map' at the end.
            if filename.endswith('.map'):
                hashed_filename = filename[:-4]
            else:
                hashed_filename = filename
            # Extract the file hash from the name
            filename_match = re.match(r'(.+)\.([0-9a-fA-F]{8})(\.\w+)$',
                    hashed_filename)
            if filename_match is None:
                current_app.logger.warning(u"Hash was unable to be found.")
                raise e
            requested_hash = filename_match.group(2)
            real_filename = filename_match.group(1) + filename_match.group(3)
            if filename != hashed_filename:
                real_filename += '.map'
            return super(VersionedStaticFlask, self).send_static_file(
                    real_filename)



from __future__ import absolute_import
from flask import url_for, redirect, abort, request, Blueprint, current_app
from flask_login import login_required, current_user
import six
from six.moves import filter, map
from sqlalchemy.orm.exc import NoResultFound
from itertools import chain

from .. import ships, systems, db
from ..models import Request, ActionType
from ..auth import PermissionType
from ..auth.models import Division, User, Group, Pilot, Entity
from .requests import PermissionRequestListing, PersonalRequests
from ..util import jsonify, classproperty


api = Blueprint('api', __name__)


filters = Blueprint('filters', __name__)


@api.route('/entities/')
@login_required
def list_entities():
    """Return a JSON object with a list of all of the specified entity type.

    Example output::
        {
          entities: [
            {name: 'Bar', id: 1, source: 'Auth Source', type: 'User'},
            {name: 'Foo', id: 0, source: 'Another Auth Source', type: 'Group'},
            {name: 'Baz', id: 20, source: 'Auth Source', type: 'Group'}
          ]
        }

    This method is only accesible to administrators.

    :param str entity_type: Either ``'user'`` or ``'group'``.
    """
    if not current_user.admin and not \
            current_user.has_permission(PermissionType.admin):
        abort(403)
    user_query = db.session.query(User.id, User.name, User.authmethod)
    group_query = db.session.query(Group.id, Group.name, Group.authmethod)
    users = map(lambda e: {
            u'id': e.id,
            u'name': e.name,
            u'type': u'User',
            u'source': e.authmethod}, user_query)
    groups = map(lambda e: {
            u'id': e.id,
            u'name': e.name,
            u'type': u'Group',
            u'source': e.authmethod}, group_query)
    return jsonify(entities=chain(users, groups))


@api.route('/user/<int:user_id>/')
def user_detail(user_id):
    user = User.query.get_or_404(user_id)
    # Set up divisions
    submit = map(lambda p: p.division,
            filter(lambda p: p.permission == PermissionType.submit,
                user.permissions))
    review = map(lambda p: p.division,
            filter(lambda p: p.permission == PermissionType.review,
                user.permissions))
    pay = map(lambda p: p.division,
            filter(lambda p: p.permission == PermissionType.pay,
                user.permissions))
    resp = {
        u'name': user.name,
        u'groups': list(user.groups),
        u'divisions': {
            u'submit': list(set(submit)),
            u'review': list(set(review)),
            u'pay': list(set(pay)),
        },
        u'admin': user.admin,
        u'requests': user.requests,
    }
    return jsonify(**resp)


@api.route('/group/<int:group_id>/')
def group_detail(group_id):
    group = Group.query.get_or_404(group_id)
    submit = map(lambda p: p.division,
            filter(lambda p: p.permission == PermissionType.submit,
                group.permissions))
    review = map(lambda p: p.division,
            filter(lambda p: p.permission == PermissionType.review,
                group.permissions))
    pay = map(lambda p: p.division,
            filter(lambda p: p.permission == PermissionType.pay,
                group.permissions))
    resp = {
        u'name': group.name,
        u'users': list(group.users),
        u'divisions': {
            u'submit': list(set(submit)),
            u'review': list(set(review)),
            u'pay': list(set(pay)),
        },
    }
    return jsonify(**resp)


@api.route('/division/')
@login_required
def list_divisions():
    """List all divisions.
    """
    if not current_user.admin:
        abort(403)
    divisions = db.session.query(Division.id, Division.name)
    return jsonify(divisions=divisions)


@api.route('/division/<int:division_id>/')
@login_required
def division_detail(division_id):
    """Get the details of a division.

    :param int division_id: The ID of the division
    """
    division = Division.query.get_or_404(division_id)
    if not current_user.admin and not \
            current_user.has_permission(PermissionType.admin, division):
        abort(403)
    permissions = {}
    for perm in PermissionType.all:
        key = perm.name + '_href'
        permissions[key] = url_for('.division_permissions',
                division_id=division_id,
                permission=perm.name)
    return jsonify(
            name=division.name,
            requests=division.requests,
            permissions=permissions)


@api.route('/division/<int:division_id>/<permission>/')
@login_required
def division_permissions(division_id, permission):
    division = Division.query.get_or_404(division_id)
    if not current_user.admin and not \
            current_user.has_permission(PermissionType.admin, division):
        abort(403)
    permission = PermissionType.from_string(permission)
    # Can't use normal Entity JSON encoder as it doesn't include the
    # authentication source or their type (explicitly. Ain't nobody got time
    # for parsing the entity type out of the href).
    entities = []
    for entity in map(lambda p: p.entity, division.permissions[permission]):
        entity_info = {
            u'name': entity.name,
            u'id': entity.id,
            u'source': str(entity.authmethod),
        }
        if hasattr(entity, u'users'):
            entity_info[u'type'] = u'Group'
            entity_info[u'length'] = len(entity.users)
        else:
            entity_info[u'type'] = u'User'
        entities.append(entity_info)
    return jsonify(
        entities=entities,
        name=permission.name,
        description=permission.description)


@api.route('/ships/')
@login_required
def ship_list():
    """Get an array of objects corresponding to every ship type.

    The objects have two keys, ``id`` is the integer typeID, and ``name`` is
    the name of the ship. This method is only accessible for logged in users to
    try to keep possible misuse to a minimum.
    """
    ship_objs = list(map(lambda s: {u'name': s[1], u'id': s[0]},
            ships.ships.items()))
    return jsonify(ships=ship_objs)


class FiltersRequestListing(object):
    @classproperty
    def _load_options(self):
        """Returns a sequence of
        :py:class:`~sqlalchemy.orm.strategy_options.Load` objects specifying
        which attributes to load.
        """
        return (
                db.Load(Request).load_only(
                    'id',
                    'pilot_id',
                    'corporation',
                    'alliance',
                    'ship_type',
                    'status',
                    'base_payout',
                    'kill_timestamp',
                    'timestamp',
                    'division_id',
                    'submitter_id',
                    'system',
                ),
                db.Load(Division).joinedload('name'),
                db.Load(Pilot).joinedload('name'),
                db.Load(User).joinedload('id')
        )


    def dispatch_request(self, filters='', **kwargs):
        def request_dict(request):
            payout = request.payout
            return {
                u'id': request.id,
                u'href': url_for('requests.get_request_details',
                    request_id=request.id),
                u'pilot': request.pilot.name,
                u'corporation': request.corporation,
                u'alliance': request.alliance,
                u'ship': request.ship_type,
                u'status': request.status.name,
                u'payout': payout.currency(),
                u'kill_timestamp': request.kill_timestamp,
                u'submit_timestamp': request.timestamp,
                u'division': request.division.name,
                u'submitter_id': request.submitter.id,
                u'system': request.system,
                u'constellation': request.constellation,
                u'region': request.region,
            }

        return jsonify(requests=map(request_dict, self.requests({})))


class APIRequestListing(FiltersRequestListing, PermissionRequestListing): pass


class APIPersonalRequests(FiltersRequestListing, PersonalRequests): pass


@filters.record
def register_request_lists(state):
    # Create the views
    all_requests = APIRequestListing.as_view('filter_requests_all',
            PermissionType.all, ActionType.statuses)
    user_requests = APIPersonalRequests.as_view('filter_requests_own')
    pending_requests = APIRequestListing.as_view('filter_requests_pending',
            (PermissionType.review,), ActionType.pending)
    pay_requests = APIRequestListing.as_view('filter_requests_pay',
            (PermissionType.pay,), (ActionType.approved,))
    completed_requests = APIRequestListing.as_view('filter_requests_completed',
            PermissionType.elevated, ActionType.finalized)
    # Attach the views to paths
    for prefix in state.app.request_prefixes:
        state.add_url_rule(prefix + '/', view_func=all_requests)
        state.add_url_rule(prefix + '/<path:filters>/',
                view_func=all_requests)
        state.add_url_rule(prefix + '/personal/', view_func=user_requests)
        state.add_url_rule(prefix + '/personal/<path:filters>/',
                view_func=user_requests)
        state.add_url_rule(prefix + '/pending/', view_func=pending_requests)
        state.add_url_rule(prefix + '/pending/<path:filters>/',
                view_func=pending_requests)
        state.add_url_rule(prefix + '/pay/', view_func=pay_requests)
        state.add_url_rule(prefix + '/pay/<path:filters>/',
                view_func=pay_requests)
        state.add_url_rule(prefix + '/completed/',
                view_func=completed_requests)
        state.add_url_rule(prefix + '/completed/<path:filters>/',
                view_func=completed_requests)


def _first(o):
    return o[0]


@filters.route('/ship/')
@login_required
def filter_ships():
    ships = db.session.query(Request.ship_type).distinct()
    return jsonify(key=u'ship', ship=map(_first, ships))


@filters.route('/system/')
@login_required
def filter_systems():
    systems = db.session.query(Request.system).distinct()
    return jsonify(key=u'system', system=map(_first, systems))


@filters.route('/constellation/')
@login_required
def filter_constellations():
    constellations = db.session.query(Request.constellation).distinct()
    return jsonify(key=u'constellation',
            constellation=map(_first, constellations))


@filters.route('/region/')
@login_required
def filter_regions():
    regions = db.session.query(Request.region).distinct()
    return jsonify(key=u'region', region=map(_first, regions))


@filters.route('/details/<path:query>')
@login_required
def query_details(query):
    requests = db.session.query(Request.id)\
            .filter(Request.details.match(query))
    return jsonify(ids=map(_first, requests))


@filters.route('/pilot/')
@login_required
def filter_pilots():
    pilots = db.session.query(Pilot.name)
    return jsonify(key=u'pilot', pilot=map(_first, pilots))


@filters.route('/corporation/')
@login_required
def filter_corps():
    corps = db.session.query(Request.corporation).distinct()
    return jsonify(key=u'corporation', corporation=map(_first, corps))


@filters.route('/alliance/')
@login_required
def filter_alliances():
    alliances = db.session.query(Request.alliance)\
            .filter(Request.alliance != None)\
            .distinct()
    return jsonify(key=u'alliance', alliance=map(_first, alliances))


@filters.route('/division/')
@login_required
def filter_divisions():
    div_names = db.session.query(Division.name)
    return jsonify(key=u'division', division=map(_first, div_names))



from __future__ import absolute_import
from __future__ import unicode_literals
from .killmail import ESIMail, ZKillmail

CSRF_ENABLE = True

# This really needs to be set to something
SRP_AUTH_METHODS = []

# Default database is an in-memory SQLite DB.
SQLALCHEMY_DATABASE_URI = 'sqlite://'

# We don't track modifications, and Flask-SQLAlchemy emits a warning if we
# don't silence it.
SQLALCHEMY_TRACK_MODIFICATIONS = False

# ESI killmails don't have any external dependencies
SRP_KILLMAIL_SOURCES = [ZKillmail, ESIMail]

SRP_SITE_NAME = 'EVE-SRP'

SRP_DETAILS_DESCRIPTION = 'Supporting details about your loss.'

SRP_SKIP_VALIDATION = False

# Add a hash of the files contents to the filename (useful for working around
# caching issues).
SRP_STATIC_FILE_HASH = False

# Currently, just show the two English locales until other translations are
# actually finished/looked at by people who can use them.
SRP_LOCALES = ['en_US', 'en_GB']

BABEL_DEFAULT_LOCALE = 'en_US'

SENTRY_USER_ATTRS = ['name', 'authmethod']
